import copy
import json
from abc import ABCMeta, abstractmethod
from typing import Optional, Tuple

import requests
from loguru import logger

from ssrspeed.config import ssrconfig
from ssrspeed.parser.conf import shadowsocks_get_config
from ssrspeed.parser.filter import NodeFilter
from ssrspeed.util import b64plus

LOCAL_ADDRESS = ssrconfig.get("localAddress", "127.0.0.1")
LOCAL_PORT = ssrconfig.get("localPort", 7890)
TIMEOUT = 10


class BottomParser(NodeFilter, metaclass=ABCMeta):
    def __init__(self):
        super().__init__()
        self._config_list: list = []
        self.__base_shadowsocks_config: dict = shadowsocks_get_config()
        self.__base_shadowsocks_config["timeout"]: int = TIMEOUT
        self.__base_shadowsocks_config["local_port"]: int = LOCAL_PORT
        self.__base_shadowsocks_config["local_address"]: str = LOCAL_ADDRESS

    def clean_configs(self):
        self._config_list = []

    def add_configs(self, new_configs: dict):
        for item in new_configs:
            self._config_list.append(item)

    @abstractmethod
    def _parse_link(self, link: str) -> Optional[dict]:
        raise NotImplementedError

    def parse_single_link(self, link: str) -> Optional[dict]:
        return self._parse_link(link)

    @staticmethod
    def _get_local_config() -> Tuple[str, int]:
        return LOCAL_ADDRESS, LOCAL_PORT

    def _get_shadowsocks_base_config(self) -> dict:
        return copy.deepcopy(self.__base_shadowsocks_config)

    def print_node(self):
        for item in self._config_list:
            logger.info(f'{item["group"]} - {item["remarks"]}')

    def read_subscription_config(self, url: str):
        header = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36"
        }
        rep = requests.get(url, headers=header, timeout=15)
        rep.encoding = "utf-8"
        res = rep.content.decode("utf-8").strip()
        links_arr = (b64plus.decode(res).decode("utf-8")).split("\n")
        for link in links_arr:
            link = link.strip()
            if cfg := self._parse_link(link):
                self._config_list.append(cfg)
        logger.info(f"Read {len(self._config_list)} node(s)")

    def read_gui_config(self, filename: str):
        with open(filename, "r", encoding="utf-8") as f:
            for item in json.load(f)["configs"]:
                _dict = {
                    "server": item["server"],
                    "server_port": int(item["server_port"]),
                    "password": item["password"],
                    "method": item["method"],
                    "protocol": item.get("protocol", ""),
                    "protocol_param": item.get("protocolparam", ""),
                    "plugin": item.get("plugin", ""),
                    "obfs": item.get("obfs", ""),
                    "obfs_param": item.get("obfsparam", ""),
                    "plugin_opts": item.get("plugin_opts", ""),
                    "plugin_args": item.get("plugin_args", ""),
                    "remarks": item.get("remarks", item["server"]),
                    "group": item.get("group", "N/A"),
                    "local_address": LOCAL_ADDRESS,
                    "local_port": LOCAL_PORT,
                    "timeout": TIMEOUT,
                    "fast_open": False,
                }
                if _dict["remarks"] == "":
                    _dict["remarks"] = _dict["server"]
                self._config_list.append(_dict)

        logger.info(f"Read {len(self._config_list)} node(s).")

    def get_all_config(self) -> list:
        return self._config_list

    def get_nodes_config(self) -> list:
        return [node.config for node in self.__node_list]

    def get_next_config(self) -> Optional[list]:
        return self._config_list.pop(0) if self._config_list else None



import asyncio
import socket
import sys

from loguru import logger


async def async_check_port(port):
    try:
        loop = asyncio.get_running_loop()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(3)
        await loop.sock_connect(sock, ("127.0.0.1", port))
        sock.shutdown(2)
        logger.info("Port Available.")
        return True
    except socket.timeout:
        logger.error(f"Port {port} timeout.")
        return False
    except ConnectionRefusedError:
        logger.error(f"Connection refused on port {port}.")
        return False
    except Exception as error:
        logger.error(f"Other Error `{error}` on port {port}.")
        return False
    finally:
        await asyncio.sleep(1)


def sync_check_port(port: int):
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(3)
        sock.connect(("127.0.0.1", port))
        sock.shutdown(2)
    except Exception:
        logger.error(f"Port {port} already in use, ")
        logger.error(
            "please change the local port in ssrspeed.json or terminate the application."
        )
        sys.exit(1)



import json
import urllib.parse

from flask import request

from ssrspeed.util.web.parseqsplus import parse_qs_plus


def get_post_data() -> dict:
    # print(request.content_type)
    data: dict = {}
    if request.content_type.startswith("application/json"):
        _data = request.get_data()
        return json.loads(_data.decode("utf-8"))
    if request.content_type.startswith("application/x-www-form-urlencoded"):
        # print(urllib.parse.parse_qs(request.get_data().decode("utf-8")))
        return parse_qs_plus(urllib.parse.parse_qs(request.get_data().decode("utf-8")))
    for key, value in request.form.items():
        if key.endswith("[]"):
            data[key[:-2]] = request.form.getlist(key)
        else:
            data[key] = value
    return data



import requests
from loguru import logger

from ssrspeed.parser.bottom import BottomParser
from ssrspeed.parser.ss import (
    ParserShadowsocksBasic,
    ParserShadowsocksClash,
    ParserShadowsocksD,
    ParserShadowsocksSIP002,
)
from ssrspeed.util import b64plus


class ShadowsocksParser(BottomParser):
    def _parse_link(self, link: str) -> dict:
        pssb = ParserShadowsocksBasic(self._get_shadowsocks_base_config())
        return pssb.parse_subs_config([link])[0]

    def read_subscription_config(self, url: str):
        header = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36"
        }
        rep = requests.get(url, headers=header, timeout=15)
        rep.encoding = "utf-8"
        res = rep.content.decode("utf-8")
        if res[:6] == "ssd://":
            logger.info("Try ShadowsocksD Parser.")
            pssd = ParserShadowsocksD(self._get_shadowsocks_base_config())
            self._config_list = pssd.parse_subs_config(
                b64plus.decode(res[6:]).decode("utf-8")
            )
        else:
            try:
                logger.info("Try Shadowsocks Basic Parser.")
                links_arr = (b64plus.decode(res).decode("utf-8")).split("\n")
                try:
                    pssb = ParserShadowsocksBasic(self._get_shadowsocks_base_config())
                    self._config_list = pssb.parse_subs_config(links_arr)
                except ValueError:
                    logger.info("Try Shadowsocks SIP002 Parser.")
                    pssip002 = ParserShadowsocksSIP002(
                        self._get_shadowsocks_base_config()
                    )
                    self._config_list = pssip002.parse_subs_config(links_arr)
            except ValueError:
                logger.info("Try Shadowsocks Clash Parser.")
                pssc = ParserShadowsocksClash(self._get_shadowsocks_base_config())
                self._config_list = pssc.parse_subs_config(res)
        logger.info(f"Read {len(self._config_list)} node(s).")

    def read_gui_config(self, filename: str):
        logger.info("Try Shadowsocks Basic or ShadowsocksD Parser.")
        pssb = ParserShadowsocksBasic(self._get_shadowsocks_base_config())
        cfg = pssb.parse_gui_config(filename)
        if not cfg:
            logger.info("Not ShadowsocksBasic or ShadowsocksD Config.")
            logger.info("Try Shadowsocks Clash Parser.")
            pssc = ParserShadowsocksClash(self._get_shadowsocks_base_config())
            cfg = pssc.parse_gui_config(filename)
        if not cfg:
            logger.info("Not Clash Configs.")
            cfg = []
            logger.critical("Unsupported config file.")
        self._config_list = cfg
        logger.info(f"Read {len(self._config_list)} node(s).")



import copy
import json

from loguru import logger


class ParserShadowsocksD:
    def __init__(self, base_config):
        self.__config_list: list = []
        self.__base_config = base_config

    def __get_shadowsocks_base_config(self):
        return copy.deepcopy(self.__base_config)

    def parse_subs_config(self, config) -> list:
        ssd_config = json.loads(config)
        group = ssd_config.get("airport", "N/A")
        default_port = int(ssd_config["port"])
        default_method = ssd_config["encryption"]
        default_password = ssd_config["password"]
        default_plugin = ssd_config.get("plugin", "")
        default_plugin_opts = ssd_config.get("plugin_options", "")
        servers = ssd_config["servers"]
        for server in servers:
            _config = self.__get_shadowsocks_base_config()
            _config["server"] = server["server"]
            _config["server_port"] = int(server.get("port", default_port))
            _config["method"] = server.get("encryption", default_method)
            _config["password"] = server.get("password", default_password)
            _config["plugin"] = server.get("plugin", default_plugin)
            _config["plugin_opts"] = server.get("plugin_options", default_plugin_opts)
            _config["group"] = group
            _config["remarks"] = server.get("remarks", server["server"])
            self.__config_list.append(_config)
        logger.info(f"Read {len(self.__config_list)} config(s).")
        return self.__config_list

    @staticmethod
    def parse_gui_config(filename: str):
        raise AttributeError(
            "'parse_gui_config' built-in 'parser_basic.py' with basic shadowsocks parser."
        )



import copy

_shadowsocks_config: dict = {
    "server": "",
    "server_port": -1,
    "method": "",
    "protocol": "",
    "obfs": "",
    "plugin": "",
    "password": "",
    "protocol_param": "",
    "obfsparam": "",
    "plugin_opts": "",
    "plugin_args": "",
    "remarks": "",
    "group": "N/A",
    "timeout": 0,
    "local_port": 0,
    "local_address": "",
    "fastopen": False,
}


def get_config(
    local_address: str = "127.0.0.1", local_port: int = 1087, timeout: int = 10
) -> dict:
    res = copy.deepcopy(_shadowsocks_config)
    res["local_port"] = local_port
    res["local_address"] = local_address
    res["timeout"] = timeout
    return res



import sys
from argparse import ArgumentParser, Namespace


def set_opts(parser: ArgumentParser):
    parser.add_argument(
        "-d",
        "--dir",
        action="store",
        dest="dir",
        type=str,
        help="Specify a work directory with clients and data.",
    )
    parser.add_argument(
        "-u",
        "--url",
        action="store",
        dest="url",
        default="",
        help="Load ssr config from subscription url.",
    )
    parser.add_argument(
        "-i",
        "--import",
        action="store",
        dest="import_file",
        default="",
        help="Import test result from json file and export it.",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        dest="guiConfig",
        default="",
        help="Load configurations from file.",
    )
    parser.add_argument(
        "-mc",
        "--max-connections",
        action="store",
        dest="max_connections",
        type=int,
        help="Set max number of connections.",
    )
    parser.add_argument(
        "-M",
        "--mode",
        action="store",
        dest="test_mode",
        default="default",
        choices=["default", "pingonly", "stream", "all", "wps"],
        help="Select test mode in [default, pingonly, stream, all, wps].",
    )
    parser.add_argument(
        "-m",
        "--method",
        action="store",
        dest="test_method",
        default="",
        choices=["stasync", "socket", "speedtestnet", "fast"],
        help="Select test method in [stasync, socket, speedtestnet, fast].",
    )
    parser.add_argument(
        "--include",
        action="extend",
        nargs="+",
        dest="filter",
        default=[],
        help="Filter nodes by group and remarks using keyword.",
    )
    parser.add_argument(
        "--include-group",
        action="extend",
        nargs="+",
        dest="group",
        default=[],
        help="Filter nodes by group name using keyword.",
    )
    parser.add_argument(
        "--include-remark",
        action="extend",
        nargs="+",
        dest="remarks",
        default=[],
        help="Filter nodes by remarks using keyword.",
    )
    parser.add_argument(
        "--exclude",
        action="extend",
        nargs="+",
        dest="efilter",
        default=[],
        help="Exclude nodes by group and remarks using keyword.",
    )
    parser.add_argument(
        "--exclude-group",
        action="extend",
        nargs="+",
        dest="egfilter",
        default=[],
        help="Exclude nodes by group using keyword.",
    )
    parser.add_argument(
        "--exclude-remark",
        action="extend",
        nargs="+",
        dest="erfilter",
        default=[],
        help="Exclude nodes by remarks using keyword.",
    )
    parser.add_argument(
        "-g",
        action="store",
        dest="group_override",
        default="",
        help="Manually set group.",
    )
    parser.add_argument(
        "-C",
        "--color",
        action="store",
        dest="result_color",
        default="origin",
        help="Set the colors when exporting images..",
    )
    parser.add_argument(
        "-s",
        "--sort",
        action="store",
        dest="sort_method",
        default="",
        choices=["speed", "rspeed", "ping", "rping"],
        help="Select sort method in [speed, rspeed, ping, rping], default not sorted.",
    )
    parser.add_argument(
        "--skip-requirements-check",
        action="store_true",
        dest="skip_requirements_check",
        default=False,
        help="Skip requirements check.",
    )
    parser.add_argument(
        "-w", "--web", action="store_true", dest="web", help="Start web server."
    )
    parser.add_argument(
        "-l",
        "--listen",
        action="store",
        dest="listen",
        help="Set listen address for web server.",
    )
    parser.add_argument(
        "-p",
        "--port",
        action="store",
        dest="port",
        type=int,
        help="Set listen port for web server.",
    )
    parser.add_argument(
        "--download",
        action="store",
        dest="download",
        choices=["all", "client", "database"],
        help="Download resources in ['all', 'client', 'database']",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        dest="debug",
        default=False,
        help="Run program in debug mode.",
    )


def init(version: str) -> Namespace:
    parser = ArgumentParser(prog="ssrspeed")
    parser.add_argument("--version", action="version", version=version)
    set_opts(parser)
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)
    return parser.parse_args()



import copy
from typing import List, Optional

from loguru import logger

from ssrspeed.util import b64plus


class ParserShadowsocksR:
    def __init__(self, base_config: dict):
        self.__base_config: dict = base_config

    def __get_base_config(self) -> dict:
        return copy.deepcopy(self.__base_config)

    def __parse_link(self, link: str) -> Optional[dict]:
        if link[:6] != "ssr://":
            logger.error(f"Unsupported link : {link}")
            return None

        decoded = b64plus.decode(link[6:]).decode("utf-8")
        decoded1 = decoded.split("/?")[0].split(":")[::-1]
        if len(decoded1) != 6:
            return None
        decoded2 = decoded.split("/?")[1].split("&")

        _config = self.__get_base_config()
        _config["server"] = decoded1[5]
        _config["server_port"] = int(decoded1[4])
        _config["method"] = decoded1[2]
        _config["password"] = b64plus.decode(decoded1[0]).decode("utf-8")
        _config["protocol"] = decoded1[3]
        _config["obfs"] = decoded1[1]

        for i in decoded2:
            if "obfsparam" in i:
                _config["obfs_param"] = b64plus.decode(i.split("=")[1]).decode("utf-8")
            elif "protocolparam" in i or "protoparam" in i:
                _config["protocol_param"] = b64plus.decode(i.split("=")[1]).decode(
                    "utf-8"
                )
            elif "remarks" in i:
                _config["remarks"] = b64plus.decode(i.split("=")[1]).decode("utf-8")
            elif "group" in i:
                _config["group"] = b64plus.decode(i.split("=")[1]).decode("utf-8")

        _config["remarks"] = _config.get("remarks", _config["server"])

        return _config

    def parse_single_link(self, link: str) -> Optional[dict]:
        return self.__parse_link(link)

    def parse_gui_data(self, data: dict) -> List[dict]:
        results = []
        for item in data["configs"]:
            _dict = self.__get_base_config()
            _dict["server"] = item["server"]
            _dict["server_port"] = int(item["server_port"])
            _dict["method"] = item["method"]
            _dict["password"] = item["password"]
            _dict["protocol"] = item.get("protocol", "")
            _dict["protocol_param"] = item.get("protocolparam", "")
            _dict["obfs"] = item.get("obfs", "")
            _dict["obfs_param"] = item.get("obfsparam", "")
            _dict["remarks"] = item.get("remarks", item["server"])
            _dict["group"] = item.get("group", "N/A")
            results.append(_dict)
        return results


if __name__ == "__main__":
    LINK = "ssr://NDUuMzIuMTMxLjExMTo4OTg5Om9yaWdpbjphZXMtMjU2LWNmYjpwbGFpbjpiM0JsYm5ObGMyRnRaUS8_cmVtYXJrcz1VMU5TVkU5UFRGOU9iMlJsT3VlLWp1V2J2U0RsaXFEbGlLbm5wb19sc0x6a3Vwcmx0NTdsbktQa3ZaWGxvWjVEYUc5dmNHSG1sYkRtamE3a3VLM2x2NE0mZ3JvdXA9VjFkWExsTlRVbFJQVDB3dVEwOU4"
    print(ParserShadowsocksR({}).parse_single_link(LINK))



import requests
from loguru import logger


def push2server(filename: str, server: str, token: str, remark: str) -> dict:
    result = {"status": -1, "code": -1}
    if not server:
        return result
    try:
        logger.info(f"Pushing {filename} to server.")
        files = {"file": open(filename, "rb")}
        param = {"token": token, "remark": remark}
        rep = requests.post(server, files=files, data=param, timeout=10)
        result["status"] = rep.status_code
        if rep.status_code == 200 and rep.text == "ok":
            result["code"] = 0
    except requests.exceptions.Timeout:
        logger.error("Connect to server timeout.")
    except Exception:
        logger.exception("Pushing result to server error.")
    return result



import asyncio
import contextlib
import copy
import time
from typing import Union

import aiohttp
from aiohttp_socks import ProxyConnector
from loguru import logger

from ssrspeed.util import ip_loc
from ssrspeed.util.rule import DownloadRuleMatch


class Statistics:
    def __init__(self):
        self._stopped: bool = False
        self._total_red: Union[int, float] = 0
        self._delta_red: Union[int, float] = 0
        self._start_time: Union[int, float] = 0
        self._statistics_time: Union[int, float] = 0
        self._time_used: Union[int, float] = 0
        self._count: int = 0
        self._speed_list: list = []

    @property
    def stopped(self) -> bool:
        return self._stopped

    @property
    def time_used(self) -> Union[int, float]:
        return self._time_used

    @property
    def total_red(self) -> Union[int, float]:
        return self._total_red

    @property
    def speed_list(self) -> list:
        return copy.deepcopy(self._speed_list)

    @property
    def max_speed(self) -> Union[int, float]:
        tmp_speed_list = self.speed_list
        tmp_speed_list.sort()
        max_speed: Union[int, float] = 0
        if len(tmp_speed_list) > 12:
            msum = 0
            for i in range(12, len(tmp_speed_list) - 2):
                msum += tmp_speed_list[i]
                max_speed = msum / (len(tmp_speed_list) - 2 - 12)
        else:
            max_speed = self._total_red / self._time_used
        return max_speed

    async def record(self, received: Union[int, float]):
        cur_time = time.time()
        if not self._start_time:
            self._start_time = cur_time
        delta_time = cur_time - self._statistics_time
        self._time_used = cur_time - self._start_time
        self._total_red += received
        if delta_time > 0.5:
            self._statistics_time = cur_time
            with contextlib.suppress(StopIteration):
                self._show_progress(delta_time)
        if self.time_used > 10:
            self._stopped = True

    def show_progress_full(self):
        mb_red = self._total_red / 1024 / 1024
        print(
            "\r["
            + "=" * self._count
            + "] ["
            + (f"{mb_red / self._time_used:.2f}" if self._time_used else "0")
            + "MB/s]"
        )
        logger.info(f"Fetched {mb_red:.2f} MB in {self._time_used:.2f}s.")

    def _show_progress(self, delta_time: Union[int, float]):
        speed = (self._total_red - self._delta_red) / delta_time
        speed_mb = speed / 1024 / 1024
        self._delta_red = self._total_red
        self._count += 1
        print("\r[" + "=" * self._count + f"> [{speed_mb:.2f} MB/s]", end="")
        self._speed_list.append(speed)


async def _fetch(url: str, sta: Statistics, host: str, port: int, buffer: int):
    try:
        logger.info(f"Fetching {url} via {host}:{port}.")
        async with aiohttp.ClientSession(
            headers={"User-Agent": "curl/11.45.14"},
            connector=ProxyConnector(host=host, port=port),
            timeout=aiohttp.ClientTimeout(connect=10),
        ) as session:
            logger.debug("Session created.")
            async with session.get(url) as response:
                logger.debug("Awaiting response.")
                while not sta.stopped:
                    chunk = await response.content.read(buffer)
                    if not chunk:
                        logger.info("No chunk, task stopped.")
                        break
                    await sta.record(len(chunk))
    except Exception as e:
        logger.error(f"Download link error: {str(e)}")


async def start(file_download: dict, proxy_host: str, proxy_port: int) -> tuple:
    download_semaphore = asyncio.Semaphore(int(file_download.get("taskNum", 1)))
    workers = file_download.get("workers", 4)
    buffer_ = file_download.get("buffer", 4096)
    async with download_semaphore:
        dlrm = DownloadRuleMatch(file_download)
        res = dlrm.get_url(await ip_loc(proxy_port))
        url = res[0]
        file_size = res[1]
        logger.debug(f"Url: {url}, file_size: {file_size} MiB.")
        logger.info(f"Running st_async, workers: {workers}.")
        _sta = Statistics()
        tasks = [
            asyncio.create_task(_fetch(url, _sta, proxy_host, proxy_port, buffer_))
            for _ in range(workers)
        ]
        await asyncio.wait(tasks)
        _sta.show_progress_full()
        if _sta.time_used:
            return (
                _sta.total_red / _sta.time_used,
                _sta.max_speed,
                _sta.speed_list,
                _sta.total_red,
            )
        return 0, 0, [], 0


if __name__ == "__main__":
    from ssrspeed.config import generate_config_file, load_path_config, ssrconfig
    from ssrspeed.path import ROOT_PATH, get_path_json

    key_path = get_path_json(ROOT_PATH)
    load_path_config({"path": key_path})
    generate_config_file()
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    logger.info(asyncio.run(start(ssrconfig["fileDownload"], "127.0.0.1", 7890)))



import asyncio
import contextlib
import copy
import os
import socket

import geoip2.database
import socks
from geoip2.errors import AddressNotFoundError
from loguru import logger

from ssrspeed.config import ssrconfig
from ssrspeed.launcher import (
    ShadowsocksClient,
    ShadowsocksRClient,
    TrojanClient,
    V2RayClient,
)
from ssrspeed.speedtest.methodology import SpeedTestMethods
from ssrspeed.util import async_check_port, domain2ip, get_ip_info, ip_loc

CLIENTS_DIR = ssrconfig["path"]["clients"]
DATABASES_DIR = ssrconfig["path"]["databases"]
TMP_DIR = ssrconfig["path"]["tmp"]
LOCAL_ADDRESS = ssrconfig.get("localAddress", "127.0.0.1")
LOCAL_PORT = int(ssrconfig.get("localPort", 10870))
MAX_CONNECTIONS = int(ssrconfig.get("maxConnections", 50))
FAST_SPEED = ssrconfig.get("fastSpeed", False)
GEOIP_TEST = ssrconfig.get("geoip", True)
STREAM_TEST = ssrconfig.get("stream", True)
PING_TEST = ssrconfig.get("ping", True)
GOOGLE_PING_TEST = ssrconfig.get("gping", True)
WPS_CONFIG = ssrconfig.get("webPageSimulation", {})
WPS_TEST = WPS_CONFIG.get("enabled", False)
FILE_DOWNLOAD = ssrconfig.get("fileDownload", {})
SPEED_TEST = ssrconfig.get("speed", False)
SOCKET_METHOD = ssrconfig.get("method", "SOCKET")
ST_SPEED_TEST = ssrconfig.get("StSpeed", False)
NTT_TEST = ssrconfig.get("ntt", {}).get("enabled", True)


class SpeedTest:
    def __init__(self, args, parser, method="ST_ASYNC"):
        self.__configs = parser.nodes
        self.__debug = args.debug
        self.__connection = args.max_connections or MAX_CONNECTIONS
        self.__test_method = method
        self.__results = []
        self.__current = {}
        self.__city_data = None
        self.__ans_data = None
        self.__base_result = {
            "group": "N/A",
            "remarks": "N/A",
            "loss": 1,
            "ping": 0,
            "gPingLoss": 1,
            "gPing": 0,
            "dspeed": -1,
            "maxDSpeed": -1,
            "trafficUsed": 0,
            "geoIP": {
                "inbound": {"address": "N/A", "info": "N/A"},
                "outbound": {"address": "N/A", "info": "N/A"},
            },
            "rawSocketSpeed": [],
            "rawTcpPingStatus": [],
            "rawGooglePingStatus": [],
            "webPageSimulation": {"results": []},
            "ntt": {
                "type": "",
                "internal_ip": "",
                "internal_port": 0,
                "public_ip": "",
                "public_port": 0,
            },
            "Ntype": "None",
            "Htype": False,
            "Dtype": False,
            "Ytype": False,
            "Atype": False,
            "Btype": False,
            "Dztype": False,
            "Ttype": False,
            "Bltype": "N/A",
            "InRes": "N/A",
            "OutRes": "N/A",
            "InIP": "N/A",
            "OutIP": "N/A",
            "port": 0,
        }
        self.__stream_cfg = {
            "NETFLIX_TEST": ssrconfig["netflix"],
            "HBO_TEST": ssrconfig["hbo"],
            "DISNEY_TEST": ssrconfig["disney"],
            "YOUTUBE_TEST": ssrconfig["youtube"],
            "ABEMA_TEST": ssrconfig["abema"],
            "BAHAMUT_TEST": ssrconfig["bahamut"],
            "DAZN_TEST": ssrconfig["dazn"],
            "TVB_TEST": ssrconfig["tvb"],
            "BILIBILI_TEST": ssrconfig["bilibili"],
        }

    def __get_base_result(self):
        return copy.deepcopy(self.__base_result)

    @staticmethod
    def __get_client(client_type, file):
        client = None
        if client_type == "Shadowsocks":
            client = ShadowsocksClient(CLIENTS_DIR, file)
        elif client_type == "ShadowsocksR":
            client = ShadowsocksRClient(CLIENTS_DIR, file)
        elif client_type == "Trojan":
            client = TrojanClient(CLIENTS_DIR, file)
        elif client_type == "V2Ray":
            client = V2RayClient(CLIENTS_DIR, file)
        return client

    def reset_status(self):
        self.__results = []
        self.__current = {}

    def get_result(self):
        return self.__results

    def get_current(self):
        return self.__current

    def load_geo_info(self):
        self.__city_data = geoip2.database.Reader(f"{DATABASES_DIR}GeoLite2-City.mmdb")
        self.__ans_data = geoip2.database.Reader(f"{DATABASES_DIR}GeoLite2-ASN.mmdb")

    def query_geo_local(self, ip):
        country, city, organization = "N/A", "Unknown City", "N/A"
        try:
            country_info = self.__city_data.city(ip).country
            country = country_info.names.get("en", "N/A")
            city = self.__city_data.city(ip).city.names.get("en", "Unknown City")
            organization = self.__ans_data.asn(ip).autonomous_system_organization
        except ValueError as e:
            logger.error(e)
        except AddressNotFoundError as e:
            logger.error(e)
        return {"country": country, "city": city, "organization": organization}

    def __geo_ip_inbound(self, config):
        inbound_ip = domain2ip(config["server"])
        inbound_geo = self.query_geo_local(inbound_ip)
        inbound_geo_res = f"{inbound_geo.get('city', 'Unknown City')}, {inbound_geo.get('organization', 'N/A')}"
        inbound_info = (
            f"{inbound_geo.get('country', 'N/A')} {inbound_geo.get('city', 'Unknown City')}, "
            f"{inbound_geo.get('organization', 'N/A')}"
        )
        return inbound_ip, inbound_geo_res, inbound_info

    @staticmethod
    async def __geo_ip_outbound(port, semaphore):
        async with semaphore:
            outbound_geo = await ip_loc(port)
        outbound_ip = outbound_geo.get("ip", "N/A")
        outbound_geo_res = f"{outbound_geo.get('city', 'Unknown City')}, {outbound_geo.get('organization', 'N/A')}"
        outbound_info = (
            f"{outbound_geo.get('country', 'N/A')} {outbound_geo.get('city', 'Unknown City')}, "
            f"{outbound_geo.get('organization', 'N/A')}"
        )
        return outbound_ip, outbound_geo_res, outbound_info

    async def __ping(self, server, server_port, address, port):
        latency_test = None
        st = SpeedTestMethods()

        res = {
            "loss": self.__base_result["loss"],
            "ping": self.__base_result["ping"],
            "rawTcpPingStatus": self.__base_result["rawTcpPingStatus"],
            "gPing": self.__base_result["gPing"],
            "gPingLoss": self.__base_result["gPingLoss"],
            "rawGooglePingStatus": self.__base_result["rawGooglePingStatus"],
        }

        if PING_TEST:
            latency_test = await st.start_tcp_ping(server, server_port)
            res["loss"] = 1 - latency_test[1]
            res["ping"] = latency_test[0]
            res["rawTcpPingStatus"] = latency_test[2]
            logger.debug(latency_test)

        if ((not PING_TEST) or (latency_test[0] > 0)) and GOOGLE_PING_TEST:
            try:
                google_ping_test = await st.start_google_ping(address, port)
                res["gPing"] = google_ping_test[0]
                res["gPingLoss"] = 1 - google_ping_test[1]
                res["rawGooglePingStatus"] = google_ping_test[2]
            except Exception:
                logger.exception("")

        return res

    @staticmethod
    def __nat_type_test(port):
        s = socks.socksocket(type=socket.SOCK_DGRAM)
        s.set_proxy(socks.PROXY_TYPE_SOCKS5, LOCAL_ADDRESS, port)
        sport = ssrconfig["ntt"]["internal_port"]
        try:
            logger.info("Performing UDP NAT Type Test.")
            t, eip, eport, sip = get_ip_info(
                source_ip=ssrconfig["ntt"]["internal_ip"],
                source_port=sport,
                include_internal=True,
                sock=s,
            )
            return t, eip, eport, sip, sport
        except (socket.gaierror, TypeError, ConnectionError) as e:
            logger.error(f"NAT Type Test: {repr(e)}")
            return None, None, None, None, None
        except Exception:
            logger.exception("NAT Type Test: ")
            return None, None, None, None, None
        finally:
            s.close()

    async def __fast_start_test(
        self, _item, cfg, address, port, geo_ip_semaphore, **kwargs
    ):
        node_info = f"[{_item['group']}] - [{_item['remarks']}] "
        geoip_log = ""
        tcp_ping_log = ""
        google_ping_log = ""
        wps_log = ""
        speed_log = ""
        nat_info = ""
        outbound_ip = "N/A"
        stream_task = None
        ping_task = None
        wps_task = None
        speed_task = None
        st = SpeedTestMethods()
        task_list = []

        if (
            kwargs.get("default", False)
            and GEOIP_TEST
            or kwargs.get("geoip_test", False)
        ):
            inbound_ip, inbound_geo_res, inbound_info = self.__geo_ip_inbound(cfg)
            (
                outbound_ip,
                outbound_geo_res,
                outbound_info,
            ) = await self.__geo_ip_outbound(port, geo_ip_semaphore)

            geoip_log = (
                f"* Inbound IP : {inbound_ip}, Geo : {inbound_info} "
                f"* Outbound IP : {outbound_ip}, Geo : {outbound_info} "
            )

            _item["geoIP"]["inbound"]["address"] = inbound_ip
            _item["InIP"] = inbound_ip
            _item["InRes"] = inbound_geo_res
            _item["geoIP"]["inbound"]["info"] = inbound_info
            _item["geoIP"]["outbound"]["address"] = outbound_ip
            _item["OutIP"] = outbound_ip
            _item["OutRes"] = outbound_geo_res
            _item["geoIP"]["outbound"]["info"] = outbound_info

        if (
            kwargs.get("default", False)
            and STREAM_TEST
            or kwargs.get("stream_test", False)
        ):
            stream_task = asyncio.create_task(
                st.start_stream_test(port, self.__stream_cfg, outbound_ip)
            )
            task_list.append(stream_task)

        if kwargs.get("default", False) and PING_TEST or kwargs.get("ping_test", False):
            ping_task = asyncio.create_task(
                self.__ping(cfg["server"], cfg["server_port"], address, port)
            )
            task_list.append(ping_task)

        if kwargs.get("default", False) and WPS_TEST or kwargs.get("wps_test", False):
            wps_task = asyncio.create_task(st.start_wps_test(WPS_CONFIG, address, port))
            task_list.append(wps_task)

        if (
            kwargs.get("default", False)
            and SPEED_TEST
            or kwargs.get("speed_test", False)
        ):
            speed_task = asyncio.create_task(
                st.start_test(
                    address=address,
                    port=port,
                    file_download=FILE_DOWNLOAD,
                    speed_test=SPEED_TEST,
                    method=self.__test_method,
                    socket_method=SOCKET_METHOD,
                    st_speed_test=ST_SPEED_TEST,
                )
            )
            task_list.append(speed_task)

        if kwargs.get("default", False) and NTT_TEST or kwargs.get("ntt_test", False):
            t, eip, eport, sip, sport = self.__nat_type_test(port)
            _item["ntt"]["type"] = t
            _item["ntt"]["internal_ip"] = sip
            _item["ntt"]["internal_port"] = sport
            _item["ntt"]["public_ip"] = eip
            _item["ntt"]["public_port"] = eport
            if t:
                nat_info = f"* NAT Type: {t} "
                if t != "Blocked":
                    nat_info += f"* Internal End: {sip}:{sport} "
                    nat_info += f"* Public End: {eip}:{eport} "

        await asyncio.wait(task_list)

        if stream_task:
            result = stream_task.result()
            _item.update(result)

        if ping_task:
            ping_res = ping_task.result()
            tcp_ping_log = (
                f"* Loss: [{ping_res['loss'] * 100:.2f}%] "
                f"* TCP Ping: [{ping_res['ping'] * 1000:.2f}] "
            )
            google_ping_log = (
                f"* Loss: [{ping_res['gPingLoss'] * 100:.2f}%] "
                f"* Google Ping: [{ping_res['gPing'] * 1000:.2f}] "
            )
            _item["ping"] = ping_res["ping"]
            _item["loss"] = ping_res["loss"]
            _item["rawTcpPingStatus"] = ping_res["rawTcpPingStatus"]
            _item["gPing"] = ping_res["gPing"]
            _item["gPingLoss"] = ping_res["gPingLoss"]
            _item["rawGooglePingStatus"] = ping_res["rawGooglePingStatus"]

        if wps_task:
            res = wps_task.result()
            wps_log = "[WebPageSimulation] "
            _item["webPageSimulation"]["results"] = res

        if speed_task:
            test_res = speed_task.result()
            if int(test_res[0]) == 0:
                logger.warning("Re-testing node.")
                test_res = await st.start_test(
                    address=address,
                    port=port,
                    file_download=FILE_DOWNLOAD,
                    speed_test=SPEED_TEST,
                    method=self.__test_method,
                    socket_method=SOCKET_METHOD,
                    st_speed_test=ST_SPEED_TEST,
                )

            speed_log = (
                f"* AvgStSpeed: [{test_res[0] / 1024 / 1024:.2f}MB/s] "
                f"* AvgMtSpeed: [{test_res[1] / 1024 / 1024:.2f}MB/s] "
            )

            _item["dspeed"] = test_res[0]
            _item["maxDSpeed"] = test_res[1]
            with contextlib.suppress(Exception):
                _item["trafficUsed"] = test_res[3]
                _item["rawSocketSpeed"] = test_res[2]
        logger.info(
            node_info
            + geoip_log
            + tcp_ping_log
            + google_ping_log
            + wps_log
            + speed_log
            + nat_info
        )

    async def __base_start_test(
        self, _item, cfg, address, port, geo_ip_semaphore, **kwargs
    ):
        node_info = f"[{_item['group']}] - [{_item['remarks']}] "
        geoip_log = ""
        tcp_ping_log = ""
        google_ping_log = ""
        wps_log = ""
        speed_log = ""
        nat_info = ""
        outbound_ip = "N/A"
        st = SpeedTestMethods()

        if (
            kwargs.get("default", False)
            and GEOIP_TEST
            or kwargs.get("geoip_test", False)
        ):
            inbound_ip, inbound_geo_res, inbound_info = self.__geo_ip_inbound(cfg)
            (
                outbound_ip,
                outbound_geo_res,
                outbound_info,
            ) = await self.__geo_ip_outbound(port, geo_ip_semaphore)

            geoip_log = (
                f"- Inbound IP : {inbound_ip}, Geo : {inbound_info} "
                f"- Outbound IP : {outbound_ip}, Geo : {outbound_info} "
            )

            _item["geoIP"]["inbound"]["address"] = inbound_ip
            _item["InIP"] = inbound_ip
            _item["InRes"] = inbound_geo_res
            _item["geoIP"]["inbound"]["info"] = inbound_info
            _item["geoIP"]["outbound"]["address"] = outbound_ip
            _item["OutIP"] = outbound_ip
            _item["OutRes"] = outbound_geo_res
            _item["geoIP"]["outbound"]["info"] = outbound_info

        if (
            kwargs.get("default", False)
            and STREAM_TEST
            or kwargs.get("stream_test", False)
        ):
            result = await st.start_stream_test(port, self.__stream_cfg, outbound_ip)
            _item.update(result)

        if (
            kwargs.get("default", False)
            and (PING_TEST or GOOGLE_PING_TEST)
            or kwargs.get("ping_test", False)
        ):
            ping_res = await self.__ping(
                cfg["server"], cfg["server_port"], address, port
            )

            tcp_ping_log = (
                f"- Loss: [{ping_res['loss'] * 100:.2f}%] "
                f"- TCP Ping: [{ping_res['ping'] * 1000:.2f}] "
            )
            google_ping_log = (
                f"- Loss: [{ping_res['gPingLoss'] * 100:.2f}%] "
                f"- Google Ping: [{ping_res['gPing'] * 1000:.2f}] "
            )

            _item["ping"] = ping_res["ping"]
            _item["loss"] = ping_res["loss"]
            _item["rawTcpPingStatus"] = ping_res["rawTcpPingStatus"]
            _item["gPing"] = ping_res["gPing"]
            _item["gPingLoss"] = ping_res["gPingLoss"]
            _item["rawGooglePingStatus"] = ping_res["rawGooglePingStatus"]

        if kwargs.get("default", False) and WPS_TEST or kwargs.get("wps_test", False):
            res = await st.start_wps_test(WPS_CONFIG, address, port)
            wps_log = "[WebPageSimulation] "
            _item["webPageSimulation"]["results"] = res

        if (
            kwargs.get("default", False)
            and SPEED_TEST
            or kwargs.get("speed_test", False)
        ):
            test_res = await st.start_test(
                address=address,
                port=port,
                file_download=FILE_DOWNLOAD,
                speed_test=SPEED_TEST,
                method=self.__test_method,
                socket_method=SOCKET_METHOD,
                st_speed_test=ST_SPEED_TEST,
            )
            if int(test_res[0]) == 0:
                logger.warning("Re-testing node.")
                test_res = await st.start_test(
                    address=address,
                    port=port,
                    file_download=FILE_DOWNLOAD,
                    speed_test=SPEED_TEST,
                    method=self.__test_method,
                    socket_method=SOCKET_METHOD,
                    st_speed_test=ST_SPEED_TEST,
                )

            speed_log = (
                f"- AvgStSpeed: [{test_res[0] / 1024 / 1024:.2f}MB/s] "
                f"- AvgMtSpeed: [{test_res[1] / 1024 / 1024:.2f}MB/s] "
            )

            _item["dspeed"] = test_res[0]
            _item["maxDSpeed"] = test_res[1]
            with contextlib.suppress(Exception):
                _item["trafficUsed"] = test_res[3]
                _item["rawSocketSpeed"] = test_res[2]
        if kwargs.get("default", False) and NTT_TEST or kwargs.get("ntt_test", False):
            t, eip, eport, sip, sport = self.__nat_type_test(port)
            _item["ntt"]["type"] = t
            _item["ntt"]["internal_ip"] = sip
            _item["ntt"]["internal_port"] = sport
            _item["ntt"]["public_ip"] = eip
            _item["ntt"]["public_port"] = eport
            if t:
                nat_info = f"- NAT Type: {t} "
                if t != "Blocked":
                    nat_info += f"- Internal End: {sip}:{sport} "
                    nat_info += f"- Public End: {eip}:{eport} "

        logger.info(
            node_info
            + geoip_log
            + tcp_ping_log
            + google_ping_log
            + wps_log
            + speed_log
            + nat_info
        )

    async def __async__start_test(
        self, node, dic, lock, port_queue, inner_method, geo_ip_semaphore, **kwargs
    ):
        port = await port_queue.get()
        cfg = node.config
        cfg.update({"local_port": port})
        async with lock:
            dic["done_nodes"] += 1
            logger.info(
                f"Starting test {cfg['group']} - {cfg['remarks']} [{dic['done_nodes']}/{dic['total_nodes']}]"
            )
        name = asyncio.current_task().get_name()
        file_ = f"{TMP_DIR}{name}.json"
        client = self.__get_client(node.node_type, file_)
        if not client:
            logger.warning(f"Unknown Node Type: {node.node_type}.")
            return False
        _item = self.__get_base_result()
        _item["group"] = cfg["group"]
        _item["remarks"] = cfg["remarks"]
        self.__current = _item
        cfg["server_port"] = int(cfg["server_port"])
        _item["port"] = cfg["server_port"]
        await client.start_client(cfg, self.__debug)

        # Check clients started
        if not client.check_alive():
            for _ in range(3):
                await client.start_client(cfg, self.__debug)
                if client.check_alive():
                    break
            else:
                logger.error("Failed to start clients.")
                return False
        logger.info("Client started.")

        # Check port
        if not await async_check_port(port):
            for _ in range(3):
                if await async_check_port(port):
                    break
            else:
                logger.error(f"Port {port} closed.")
                return False

        await inner_method(_item, cfg, LOCAL_ADDRESS, port, geo_ip_semaphore, **kwargs)

        self.__results.append(_item)
        port_queue.put_nowait(port)
        if client:
            client.stop_client()
        if os.path.exists(file_):
            os.remove(file_)

    async def __run(self, **kwargs):
        lock = asyncio.Lock()
        geo_ip_semaphore = asyncio.Semaphore()
        port_queue = asyncio.Queue()
        dic = {"done_nodes": 0, "total_nodes": len(self.__configs)}
        # 根据配置文件是否选择极速模式
        fast_method = self.__fast_start_test if FAST_SPEED else self.__base_start_test

        # 初始化端口范围
        for i in range(LOCAL_PORT, LOCAL_PORT + self.__connection):
            port_queue.put_nowait(i)

        # 布置异步任务
        task_list = [
            asyncio.create_task(
                self.__async__start_test(
                    node, dic, lock, port_queue, fast_method, geo_ip_semaphore, **kwargs
                )
            )
            for node in self.__configs
        ]

        await asyncio.wait(task_list)

    def __start_test(self, **kwargs):
        self.__results = []
        self.load_geo_info()
        if os.name == "nt":
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.__run(**kwargs))
        loop.close()
        self.__current = {}

    def default_test(self):
        logger.info("Test Default")
        self.__start_test(default=True)

    def web_page_simulation(self):
        logger.info("Test mode : Web Page Simulation")
        self.__start_test(geoip_test=True, ping_test=True, wps_test=True)

    def ping_only(self):
        logger.info("Test mode : ping only")
        self.__start_test(ping_test=True)

    def stream_only(self):
        logger.info("Test mode : stream only")
        self.__start_test(stream_test=True)

    def full_test(self):
        logger.info(f"Test mode : All. Test method : {self.__test_method}")
        self.__start_test(
            geoip_test=True,
            stream_test=True,
            ping_test=True,
            ntt_test=True,
            speed_test=True,
        )



from .basic import ParserShadowsocksBasic
from .sclash import ParserShadowsocksClash
from .sip002 import ParserShadowsocksSIP002
from .ssd import ParserShadowsocksD

__all__ = [
    "ParserShadowsocksBasic",
    "ParserShadowsocksClash",
    "ParserShadowsocksSIP002",
    "ParserShadowsocksD",
]



# PyNAT: Discover external IP addresses and NAT topologies using STUN.
# Copyright (C) 2022 Ariel A. Licensed under the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# pynat.py
"""PyNAT v0.7.0

Discover external IP addresses and NAT topologies using STUN.

Copyright (C) 2022 Ariel A. Licensed under the MIT License.

Drop Python 2 support by Oreomeow, 2022-09-12.
"""
import argparse
import codecs
import ipaddress
import random
import secrets
import socket
import sys

__version__ = "0.7.0"
URL = "https://github.com/aarant/pynat"


class PynatError(Exception):
    """Raised when an error occurs during network discovery."""


# Non-NAT network topologies
BLOCKED = "Blocked"
OPEN = "Open"
UDP_FIREWALL = "UDP Firewall"
# NAT topologies
FULL_CONE = "Full-cone NAT"
RESTRICTED_CONE = "Restricted-cone NAT"
RESTRICTED_PORT = "Restricted-port NAT"
SYMMETRIC = "Symmetric NAT"

# Stun message types
BIND_REQUEST_MSG = b"\x00\x01"
BIND_RESPONSE_MSG = b"\x01\x01"
MAGIC_COOKIE = b"\x21\x12\xA4\x42"

# Stun attributes
MAPPED_ADDRESS = b"\x00\x01"
RESPONSE_ADDRESS = b"\x00\x02"
CHANGE_REQUEST = b"\x00\x03"
SOURCE_ADDRESS = b"\x00\x04"
CHANGED_ADDRESS = b"\x00\x05"
XOR_MAPPED_ADDRESS = b"\x00\x20"

# List of classic STUN servers
STUN_SERVERS = [
    ("stun.ekiga.net", 3478),
    ("stun.ideasip.com", 3478),
    ("stun.voiparound.com", 3478),
    ("stun.voipbuster.com", 3478),
    ("stun.voipstunt.com", 3478),
    ("stun.voxgratia.org", 3478),
]


def randint(n):
    return secrets.randbits(n)


def ord_(ch):  # compatible to python3
    return ch if isinstance(ch, int) else ord(ch)


# Get the family of an IP address
def get_address_family(addr):
    try:
        ipaddress.IPv4Interface(addr)
        return socket.AF_INET
    except ipaddress.AddressValueError:
        try:
            ipaddress.IPv6Interface(addr)
            return socket.AF_INET6
        except ipaddress.AddressValueError as e:
            raise PynatError(f"Invalid IP address: {addr}") from e


# Send a STUN message to a server, with optional extra data
def send_stun_message(sock, addr, msg_type, trans_id=None, send_data=b""):
    if trans_id is None:
        trans_id = randint(128).to_bytes(16, byteorder="big")
    msg_len = len(send_data).to_bytes(2, byteorder="big")
    data = msg_type + msg_len + trans_id + send_data
    sock.sendto(data, addr)
    return trans_id


# Get a STUN Binding response from a server, with optional extra data
def get_stun_response(sock, addr, trans_id=None, send_data=b"", max_timeouts=6):
    timeouts = 0
    response = None
    old_timeout = sock.gettimeout()
    sock.settimeout(0.5)
    while timeouts < max_timeouts:
        try:
            trans_id = send_stun_message(
                sock, addr, BIND_REQUEST_MSG, trans_id, send_data
            )
            recv, addr = sock.recvfrom(2048)  # TODO: Why 2048
        except socket.timeout:
            timeouts += 1
            continue
        else:
            # Too short, not a valid message
            if len(recv) < 20:
                continue
            msg_type, recv_trans_id, attrs = recv[:2], recv[4:20], recv[20:]
            msg_len = int(codecs.encode(recv[2:4], "hex"), 16)
            if msg_len != len(attrs):
                continue
            if msg_type != BIND_RESPONSE_MSG:
                continue
            if recv_trans_id != trans_id:
                continue
            response = {}
            i = 0
            while i < msg_len:
                attr_type, attr_length = attrs[i : i + 2], int(
                    codecs.encode(attrs[i + 2 : i + 4], "hex"), 16
                )
                attr_value = attrs[i + 4 : i + 4 + attr_length]
                i += 4 + attr_length
                if (
                    attr_length % 4 != 0
                ):  # If not on a 32-bit boundary, add padding bytes
                    i += 4 - (attr_length % 4)
                if attr_type in [MAPPED_ADDRESS, SOURCE_ADDRESS, CHANGED_ADDRESS]:
                    family, port = ord_(attr_value[1]), int(
                        codecs.encode(attr_value[2:4], "hex"), 16
                    )
                    if family == 0x01:  # IPv4
                        ip = socket.inet_ntop(socket.AF_INET, attr_value[4:8])
                        if attr_type == XOR_MAPPED_ADDRESS:
                            cookie_int = int(codecs.encode(MAGIC_COOKIE, "hex"), 16)
                            port ^= cookie_int >> 16
                            ip = (
                                int(codecs.encode(attr_value[4:8], "hex"), 16)
                                ^ cookie_int
                            )
                            ip = socket.inet_ntoa(ip.to_bytes(4, byteorder="big"))
                            response["xor_ip"], response["xor_port"] = ip, port
                        elif attr_type == MAPPED_ADDRESS:
                            response["ext_ip"], response["ext_port"] = ip, port
                        elif attr_type == SOURCE_ADDRESS:
                            response["src_ip"], response["src_port"] = ip, port
                        elif attr_type == CHANGED_ADDRESS:
                            response["change_ip"], response["change_port"] = (ip, port)
                    else:  # family == 0x02:  # IPv6
                        ip = socket.inet_ntop(socket.AF_INET6, attr_value[4:20])
                        if attr_type == XOR_MAPPED_ADDRESS:
                            cookie_int = int(codecs.encode(MAGIC_COOKIE, "hex"), 16)
                            port ^= cookie_int >> 16
                            ip = int(codecs.encode(attr_value[4:20], "hex"), 16) ^ (
                                cookie_int << 96 | trans_id
                            )
                            ip = socket.inet_ntop(
                                socket.AF_INET6, ip.to_bytes(32, byteorder="big")
                            )
                            response["xor_ip"], response["xor_port"] = ip, port
                        elif attr_type == MAPPED_ADDRESS:
                            response["ext_ip"], response["ext_port"] = ip, port
                        elif attr_type == SOURCE_ADDRESS:
                            response["src_ip"], response["src_port"] = ip, port
                        elif attr_type == CHANGED_ADDRESS:
                            response["change_ip"], response["change_port"] = (ip, port)
            # Prefer, when possible, to use XORed IPs and ports
            xor_ip, xor_port = response.get("xor_ip"), response.get("xor_port")
            if xor_ip is not None:
                response["ext_ip"] = xor_ip
            if xor_port is not None:
                response["ext_port"] = xor_port
            break
    sock.settimeout(old_timeout)
    return response


# Retrieve the internal working IPv4 used to access the Internet
def get_internal_ipv4(test_addr=("8.8.8.8", 80)):  # By default, queries Google's DNS
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.connect(test_addr)
    ip = sock.getsockname()[0]
    sock.close()
    return ip


# Retrieve the internal working IPv6 used to access the Internet
def get_internal_ipv6(
    test_addr=("2001:4860:4860::8888", 80)
):  # By default, queries Google's DNS
    sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    sock.connect(test_addr)
    ip = sock.getsockname()[0]
    sock.close()
    return ip


# Get a STUN binding response from a server, without any CHANGE_REQUEST flags
def stun_test_1(sock, addr):
    return get_stun_response(sock, addr)


# Get a STUN binding response from a server, asking it to change both the IP & port from which it replies
def stun_test_2(sock, addr):
    return get_stun_response(
        sock, addr, send_data=CHANGE_REQUEST + b"\x00\x04" + b"\x00\x00\x00\x06"
    )


# Get a STUN binding response from a server, asking it to change just the port from which it replies
def stun_test_3(sock, addr):
    return get_stun_response(
        sock, addr, send_data=CHANGE_REQUEST + b"\x00\x04" + b"\x00\x00\x00\x02"
    )


# Find a working classic STUN server from the list
def find_stun_server(sock):
    # Randomize the list so as to avoid using the same server twice in a row
    random.shuffle(STUN_SERVERS)
    for stun_addr in STUN_SERVERS:
        try:
            response = get_stun_response(sock, stun_addr, max_timeouts=1)
        except socket.gaierror:  # Host not found error
            continue
        else:
            if response is not None:  # Have found a working server
                return stun_addr
    return None


# Get the network topology, external IP, and external port
def get_ip_info(
    source_ip="0.0.0.0",
    source_port=54320,
    stun_host=None,
    stun_port=3478,
    include_internal=False,
    sock=None,
):
    """Get information about the network topology, external IP, and external port.

    Args:
        source_ip (str, optional): If not '0.0.0.0', the internal IP address to bind to. Defaults to '0.0.0.0'.
        source_port (int, optional): Source port to bind to. Defaults to 54320.
        stun_host (str, optional): Address of the STUN host to use. Defaults to None, in which case one is selected.
        stun_port (int, optional): Port of the STUN host to query. Defaults to 3478.
        include_internal (bool, optional): Whether to include internal IP address information. Defaults to False.
        sock (socket.socket, optional): Bound socket to connect with. If not provided, one will be created.

    Returns:
        tuple: (topology, external_ip, external_port). Topology & external_ip are strings, external_port is int.
            If `include_internal` is True, returns (topology, external_ip, external_port, internal_ip) instead.
    """
    # If no socket is passed in, create one and close it when done
    ephemeral_sock = sock is None
    if sock is None:
        family = get_address_family(source_ip)
        sock = socket.socket(family, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((source_ip, source_port))
    # Find a stun host if none was selected
    if stun_host is None:
        stun_addr = find_stun_server(sock)
        # If None was found, assume the network is blocked
        if stun_addr is None:
            if ephemeral_sock:
                sock.close()
            if include_internal:
                return BLOCKED, None, None, None
            return BLOCKED, None, None
    # If a stun host was specified, set stun_addr
    else:
        stun_addr = (stun_host, stun_port)
    # Determine the actual local, or source IP
    if source_ip == "0.0.0.0":
        # IPv4
        source_ip = get_internal_ipv4(stun_addr)
    elif source_ip == "::":
        # IPv6
        source_ip = get_internal_ipv6(stun_addr)
    # Perform Test 1, a simple STUN Request
    response = stun_test_1(sock, stun_addr)
    # If the test fails, assume the network blocked
    if response is None:
        if ephemeral_sock:
            sock.close()
        if include_internal:
            return BLOCKED, None, None, None
        return BLOCKED, None, None
    # Otherwise the network is not blocked and we can continue
    ext_ip, ext_port = response["ext_ip"], response["ext_port"]
    change_addr = response.get("change_ip"), response.get("change_port")
    # Either Open Internet or a UDP firewall, do test 2
    if ext_ip == source_ip and ext_port == source_port:
        response = stun_test_2(sock, stun_addr)
        # Open Internet or Symmetric UDP Firewall
        topology = OPEN if response is not None else UDP_FIREWALL
    # Some type of NAT, do test 2
    else:
        response = stun_test_2(sock, stun_addr)
        # Full-cone NAT
        if response is not None:
            topology = FULL_CONE
        # Some other type of NAT, do test 1 on a new ip
        else:
            response = stun_test_1(sock, change_addr)
            # This should never occur
            if response is None:
                if ephemeral_sock:
                    sock.close()
                raise PynatError("Error querying STUN server with changed address.")
            # Symmetric, restricted cone, or restricted port NAT
            recv_ext_ip, recv_ext_port = (response["ext_ip"], response["ext_port"])
            # Some type of restricted NAT, do test 3 to the change_addr with a CHANGE_REQUEST for the port
            if recv_ext_ip == ext_ip and recv_ext_port == ext_port:
                response = stun_test_3(sock, change_addr)
                # Restricted cone NAT or Restricted cone NAT
                topology = RESTRICTED_CONE if response is not None else RESTRICTED_PORT
            # Symmetric NAT
            else:
                topology = SYMMETRIC
    if ephemeral_sock:
        sock.close()
    if include_internal:
        return topology, ext_ip, ext_port, source_ip
    return topology, ext_ip, ext_port


def main():
    try:
        parser = argparse.ArgumentParser(prog="pynat", description=__doc__)
        parser.add_argument(
            "--source_ip",
            help="The source IPv4/IPv6 address to bind to.",
            type=str,
            default="0.0.0.0",
        )
        parser.add_argument(
            "--source-port", help="The source port to bind to.", type=int, default=54320
        )
        parser.add_argument(
            "--stun-host", help="The STUN host to use for queries.", type=str
        )
        parser.add_argument(
            "--stun-port",
            help="The port of the STUN host to use for queries.",
            type=int,
            default=3478,
        )
        args = parser.parse_args()
        source_ip, source_port, stun_host, stun_port = (
            args.source_ip,
            args.source_port,
            args.stun_host,
            args.stun_port,
        )
        topology, ext_ip, ext_port, source_ip = get_ip_info(
            source_ip, source_port, stun_host, stun_port, True
        )
        print(
            "Network type:",
            topology,
            f"\nInternal address: {source_ip}:{source_port}",
            f"\nExternal address: {ext_ip}:{ext_port}",
        )
    except KeyboardInterrupt:
        sys.exit()
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



from copy import deepcopy

from loguru import logger


class DownloadRuleMatch:
    def __init__(self, file_download: dict):
        self._config: dict = deepcopy(file_download)
        self._download_links: dict = deepcopy(file_download["downloadLinks"])
        self._download_dict: dict = {
            link["tag"]: (link["link"], link["fileSize"])
            for link in self._download_links
        }

    def _get_download_link(self, tag: str = "") -> tuple:
        if tag in self._download_dict:
            return self._download_dict[tag]
        logger.info(f"No (matched) tag {tag}, using default.")
        return self._download_dict.get("default", ("", ""))

    def _check_rule(self, data: dict) -> tuple:
        isp = data["organization"].strip()
        country_code = data["country_code"].strip()
        continent = data["continent_code"].strip()

        rules = self._config["rules"]
        for rule in rules:
            if rule["mode"].lower() == "match_isp":
                logger.debug("Match mode: ISP")
                if isp in rule["ISP"].strip():
                    logger.info(f"ISP {isp} matched.")
                    return self._get_download_link(rule["tag"])
            elif rule["mode"].lower() == "match_location":
                logger.debug("Match mode: Location")
                for code in rule.get("countries", []):
                    if country_code == code.strip():
                        logger.info(f"Country {code} matched.")
                        return self._get_download_link(rule["tag"])
                if rule.get("continent", "").strip() in continent:
                    logger.info(f"Continent {continent} matched.")
                    return self._get_download_link(rule["tag"])
        logger.info("Rule not matched, using default.")
        return self._get_download_link()

    def get_url(self, data: dict) -> tuple:
        try:
            if data and not self._config["skipRuleMatch"]:
                return self._check_rule(data)
            return self._get_download_link()
        except Exception:
            logger.exception("")
            return self._get_download_link()



import platform


def check_platform() -> str:
    tmp = platform.platform()
    # print(tmp)
    if "Windows" in tmp:
        return "Windows"
    if "Linux" in tmp:
        return "Linux"
    return "MacOS" if "Darwin" in tmp or "mac" in tmp else "Unknown"


PLATFORM = check_platform()



# Author: ranwen NyanChan
import asyncio
import socket
import time

from loguru import logger


async def tcp_ping_task(loop, _list, address, port):
    alt = fac = suc = 0
    try:
        sock = socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM)
        sock.settimeout(3)
        current_time = time.time()
        await loop.sock_connect(sock, (address, port))
        sock.close()
        delta_time = time.time() - current_time
        alt += delta_time
        suc += 1
        _list.append(delta_time)
    except socket.timeout:
        fac += 1
        _list.append(0)
        logger.error(f"TCP Ping ({address},{port}) Timeout {fac} times.")
    except ConnectionResetError:
        logger.error("TCP Ping Reset")
        _list.append(0)
        fac += 1
    except ConnectionRefusedError:
        logger.error("TCP Ping RefusedError")
        _list.append(0)
        fac += 1
    except Exception:
        logger.exception("TCP Ping Exception:")
        _list.append(0)
        fac += 1
    return {"alt": alt, "fac": fac, "suc": suc}


async def google_ping_task(loop, _list, address, port):
    alt = fac = suc = 0
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(3)
        await loop.sock_connect(sock, (address, port))
        current_time = time.time()
        await loop.sock_sendall(sock, b"\x05\x01\x00")
        await loop.sock_recv(sock, 2)
        await loop.sock_sendall(sock, b"\x05\x01\x00\x03\x0agoogle.com\x00\x50")
        await loop.sock_recv(sock, 10)
        await loop.sock_sendall(
            sock,
            b"GET / HTTP/1.1\r\nHost: google.com\r\nUser-Agent: curl/11.45.14\r\n\r\n",
        )
        await loop.sock_recv(sock, 1)
        sock.close()
        delta_time = time.time() - current_time
        alt += delta_time
        suc += 1
        _list.append(delta_time)
    except socket.timeout:
        fac += 1
        _list.append(0)
        logger.error(f"Google Ping Timeout {fac} times.")
    except Exception:
        logger.exception("Google Ping Exception:")
        _list.append(0)
        fac += 1
    return {"alt": alt, "fac": fac, "suc": suc}


async def tcp_ping(address: str, port: int) -> tuple:
    alt = fac = suc = 0
    loop = asyncio.get_running_loop()
    _list: list = []
    test_ping = [
        asyncio.create_task(tcp_ping_task(loop, _list, address, port)) for _ in range(3)
    ]
    done, _ = await asyncio.wait(test_ping, timeout=5)
    for each in done:
        result = each.result()
        alt += result["alt"]
        fac += result["fac"]
        suc += result["suc"]
    return (0, 0, _list) if suc == 0 else (alt / suc, suc / (suc + fac), _list)


async def google_ping(address: str, port: int) -> tuple:
    alt = fac = suc = 0
    loop = asyncio.get_running_loop()
    _list: list = []
    test_ping = [
        asyncio.create_task(google_ping_task(loop, _list, address, port))
        for _ in range(3)
    ]
    done, _ = await asyncio.wait(test_ping, timeout=5)
    for each in done:
        result = each.result()
        alt += result["alt"]
        fac += result["fac"]
        suc += result["suc"]
    return (0, 0, _list) if suc == 0 else (alt / suc, suc / (suc + fac), _list)


if __name__ == "__main__":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    logger.info(asyncio.run(tcp_ping("127.0.0.1", 7890)))
    logger.info(asyncio.run(google_ping("127.0.0.1", 7890)))



from copy import deepcopy

from loguru import logger


class NodeFilter:
    def __init__(self):
        self.__node_list: list = []

    def filter_node(self, nodes: list, **kwargs: list) -> list:
        self.__node_list.clear()
        self.__node_list = deepcopy(nodes)
        self.__get_first_nodes(self.__node_list)
        if kwl := kwargs.get("fk", []):
            self.__filter_node(kwl)
        if gkwl := kwargs.get("frk", []):
            self.__filter_group(gkwl)
        if rkwl := kwargs.get("fgk", []):
            self.__filter_remark(rkwl)
        if ekwl := kwargs.get("ek", []):
            self.__exclude_nodes(ekwl)
        if egkwl := kwargs.get("egk", []):
            self.__exclude_group(egkwl)
        if erkwl := kwargs.get("erk", []):
            self.__exclude_remark(erkwl)
        return self.__node_list

    def __get_first_nodes(self, nodes: list):
        """
        Get the first unique node of node list_dict
        Regardless of order can use
        `list({(d.get("server", ""), d.get("port", 0)): d for d in nodes}.values())`
        """
        existing_dicts = set()
        filtered_list = []
        for node in nodes:
            d = node.config
            if (
                d.get("server", ""),
                d.get("server_port", d.get("port", 0)),
            ) not in existing_dicts:
                existing_dicts.add(
                    (d.get("server", ""), d.get("server_port", d.get("port", 0)))
                )
                filtered_list.append(node)
            else:
                logger.warning(
                    f'{d.get("group", "N/A")} - {d.get("remarks", "N/A")} '
                    f'({d.get("server", "Server EMPTY")}:{d.get("server_port", d.get("port", 0))}) '
                    f"already in list. "
                )
        self.__node_list = filtered_list

    def __filter_node(self, kwl: list):
        self.__node_list = [
            n
            for n in self.__node_list
            if any(kw in n.config["group"] or kw in n.config["remarks"] for kw in kwl)
        ]

    def __filter_group(self, gkwl: list):
        self.__node_list = [
            n for n in self.__node_list if any(kw in n.config["group"] for kw in gkwl)
        ]

    def __filter_remark(self, rkwl: list):
        self.__node_list = [
            n for n in self.__node_list if any(kw in n.config["remarks"] for kw in rkwl)
        ]

    def __exclude_nodes(self, ekwl: list):
        self.__node_list = [
            n
            for n in self.__node_list
            if all(
                kw not in n.config["group"] and kw not in n.config["remarks"]
                for kw in ekwl
            )
        ]

    def __exclude_group(self, egkwl: list):
        self.__node_list = [
            n
            for n in self.__node_list
            if all(kw not in n.config["group"] for kw in egkwl)
        ]

    def __exclude_remark(self, erkwl: list):
        self.__node_list = [
            n
            for n in self.__node_list
            if all(kw not in n.config["remarks"] for kw in erkwl)
        ]



import contextlib
from copy import deepcopy
from urllib.parse import parse_qsl, unquote, urlparse

from loguru import logger

from ssrspeed.parser.bottom import BottomParser


class TrojanParser(BottomParser):
    def __init__(self):
        super().__init__()
        self.__base_config: dict = {
            "run_type": "client",
            "local_addr": "127.0.0.1",
            "local_port": 10870,
            "remote_addr": "example.com",
            "remote_port": 443,
            "password": ["password1"],
            "log_level": 1,
            "ssl": {
                "verify": "true",
                "verify_hostname": "true",
                "cert": "",
                "cipher": "ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:"
                "ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:"
                "ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384:"
                "ECDHE-ECDSA-AES256-SHA:ECDHE-ECDSA-AES128-SHA:"
                "ECDHE-RSA-AES128-SHA:ECDHE-RSA-AES256-SHA:"
                "DHE-RSA-AES128-SHA:DHE-RSA-AES256-SHA:"
                "AES128-SHA:AES256-SHA:DES-CBC3-SHA",
                "cipher_tls13": "TLS_AES_128_GCM_SHA256:TLS_CHACHA20_POLY1305_SHA256:TLS_AES_256_GCM_SHA384",
                "sni": "",
                "alpn": ["h2", "http/1.1"],
                "reuse_session": "true",
                "session_ticket": "false",
                "curves": "",
            },
            "tcp": {
                "no_delay": "true",
                "keep_alive": "true",
                "reuse_port": "false",
                "fast_open": "false",
                "fast_open_qlen": 20,
            },
            "websocket": {"enabled": "false", "path": "", "host": ""},
            "group": "N/A",
        }

    def __get_trojan_base_config(self) -> dict:
        return deepcopy(self.__base_config)

    @staticmethod
    def percent_decode(s: str) -> str:
        try:
            s = unquote(s, encoding="gb2312", errors="strict")
        except Exception:
            with contextlib.suppress(Exception):
                s = unquote(s, encoding="utf8", errors="strict")
        return s

    def _parse_link(self, link: str) -> dict:
        url = urlparse(link.strip("\n"))
        if not url.scheme.startswith("trojan"):
            logger.error(f"Not trojan URL : {link}")
            return {}

        _config = self.__get_trojan_base_config()
        try:
            hostname = url.hostname
            port = url.port or 443
            password = unquote(url.netloc.split("@")[0])
            query = dict(parse_qsl(url.query))
            remarks = unquote(url.fragment)

            _config["remote_addr"], _config["server"] = hostname, hostname
            _config["remote_port"], _config["server_port"] = port, port
            _config["password"][0] = password
            _config["remarks"] = remarks
            _config["ssl"]["verify"] = query.get("allowinsecure", "") == "1"
            _config["ssl"]["sni"] = query.get("sni", "")
            _config["tcp"]["fast_open"] = query.get("tfo", "") == "1"
            _config["group"] = query.get("peer", "N/A")

            # ws protocol must pass host and path
            if query.get("type", "") == "ws":
                _config["websocket"]["enabled"] = "true"
                _config["websocket"]["path"] = query.get("path", "")
                _config["websocket"]["host"] = query.get("host", "")

        except Exception:
            logger.exception(f"Invalid trojan URL : {link}")
        return _config


if __name__ == "__main__":
    LINKS = (
        "trojan://8888@[2001:1234:4321:66::33]?allowinsecure=0&tfo=1\n"
        "trojan://123@helloworld.xyz?allowinsecure=0&tfo=1#testIPv4port443\n"
        "trojan://9999@[2001:1234:4321:66::33]:444?allowinsecure=1&tfo=0#testIPv6\n"
        "trojan://1234@a.helloworld.xyz:444?allowinsecure=1&tfo=0#%E4%BD%A0%E5%A5%BD\n"
        "trojan://12345@a.b.c:445?security=tls&sni=a.b.c&alpn=http%2F1.1&type=tcp&headerType=none\n"
        "trojan-go://1@ip.com:446/?sni=qq.com&type=ws&host=fast.com&path=%2Fgo&encryption=ss%3Baes-256-gcm%3Afuckgfw\n"
    )
    tropar = TrojanParser()
    for i in LINKS.split("\n"):
        if i:
            print(tropar.parse_single_link(i))



class Sorter:
    @staticmethod
    def __sort_by_speed(result: dict) -> float:
        return result["dspeed"]

    @staticmethod
    def __sort_by_ping(result: dict) -> float:
        return result["ping"]

    def sort_result(self, result: list, sort_method: str) -> list:
        if sort_method == "PING":
            result.sort(key=self.__sort_by_ping)
        elif sort_method == "REVERSE_PING":
            result.sort(key=self.__sort_by_ping, reverse=True)
        elif sort_method == "REVERSE_SPEED":
            result.sort(key=self.__sort_by_speed)
        elif sort_method == "SPEED":
            result.sort(key=self.__sort_by_speed, reverse=True)
        return result



from copy import deepcopy


class BasisNode:
    def __init__(self, config: dict):
        self._type: str = ""
        if not isinstance(config, dict):
            raise TypeError("The 'config' parameter must be dict")
        self._config: dict = config

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>, Proxy Type: {self.node_type}"

    @property
    def node_type(self) -> str:
        if not self._type:
            raise NotImplementedError
        return self._type

    @property
    def config(self) -> dict:
        return deepcopy(self._config)

    def update_config(self, new_cfg: dict):
        if new_cfg:
            self._config.update(new_cfg)

    def __key(self):
        return self._config["server"], self._config["server_port"]

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other) -> bool:
        if isinstance(other, BasisNode):
            return self.__key() == other.__key()
        return NotImplemented



import json
import os
import shutil
from typing import Any, Dict

config: Dict[str, Any] = {}


def load_path_config(data: dict):
    config.update(data)


def mkdir(data_dir: str):
    if os.path.exists(data_dir):
        if os.path.isfile(data_dir):
            os.remove(data_dir)
            os.makedirs(data_dir)
    else:
        os.makedirs(data_dir)


def generate_config_file():
    data_dir = config["path"]["data"]
    config_file = config["path"]["ssrspeed.json"]
    config_example_file = config["path"]["ssrspeed.example.json"]
    mkdir(data_dir)
    if os.path.exists(config_file):
        if os.path.isdir(config_file):
            shutil.rmtree(config_file)
            if not os.path.exists(config_example_file):
                raise FileNotFoundError(
                    "Default configuration file not found, please download from the official repo and try again."
                )
            shutil.copy(config_example_file, config_file)
    else:
        if not os.path.exists(config_example_file):
            raise FileNotFoundError(
                "Default configuration file not found, please download from the official repo and try again."
            )
        shutil.copy(config_example_file, config_file)
    with open(config_file, "r", encoding="utf-8") as f:
        file_config = json.load(f)
        config.update(file_config)



"""
Python CLI-tool (without need for a GUI) to measure Internet speed with fast.com
"""


import contextlib
import json
import socket
import time
import urllib.error
import urllib.request
from threading import Thread

import socks
from loguru import logger


def set_proxy(local_address, local_port):
    socks.setdefaultproxy(socks.PROXY_TYPE_SOCKS5, local_address, local_port)
    socket.socket = socks.socksocket


def get_html_result(url, result, index):
    """
    get the stuff from url in chunks of size CHUNK,
    and keep writing the number of bytes retrieved into result[index]
    """
    try:
        req = urllib.request.urlopen(url)
    except urllib.error.URLError:
        result[index] = 0
        return
    CHUNK = 100 * 1024
    i = 1
    while True:
        chunk = req.read(CHUNK)
        if not chunk:
            break
        result[index] = i * CHUNK
        i += 1


def application_bytes_to_networkbits(bytes_):
    # convert bytes (at application layer) to bits (at network layer)
    return bytes_ * 8 * 1.0415
    # 8 for bits versus bytes
    # 1.0416 for application versus network layers


def find_ipv4(fqdn):
    """find IPv4 address of fqdn"""
    return socket.getaddrinfo(fqdn, 80, socket.AF_INET)[0][4][0]


def find_ipv6(fqdn):
    """find IPv6 address of fqdn"""
    return socket.getaddrinfo(fqdn, 80, socket.AF_INET6)[0][4][0]


def fast_com(verbose=False, max_time=15, force_ipv4=False, force_ipv6=False):
    """
    verbose: print debug output
    max_time: max time in seconds to monitor speedtest
    force_ipv4: force speed test over IPv4
    force_ipv6: force speed test over IPv6
    """
    # go to fast.com to get the javascript file
    url = "https://fast.com/"
    try:
        url_result = urllib.request.urlopen(url)
    except Exception:
        logger.exception("No connection at all")
        # no connection at all?
        return 0
    response = url_result.read().decode().strip()
    for line in response.split("\n"):
        # We're looking for a line like
        #           <script ssrspeed="/app-40647a.js"></script>
        if line.find("script src") >= 0:
            jsname = line.split('"')[1]  # At time of writing: '/app-40647a.js'

    # From that javascript file, get the token:
    url = f"https://fast.com{jsname}"
    if verbose:
        logger.debug(f"javascript url is{url}")
    try:
        url_result = urllib.request.urlopen(url)
    except Exception:
        # connection is broken
        return 0
    all_js_stuff = (
        url_result.read().decode().strip()
    )  # this is an obfuscated Javascript file

    # We're searching for the "token:" in this string:
    # .dummy,DEFAULT_PARAMS={https:!0,token:"YXNkZmFzZGxmbnNkYWZoYXNkZmhrYWxm",urlCount:3,e
    for line in all_js_stuff.split(","):
        if line.find("token:") >= 0:
            if verbose:
                logger.debug(f"line is{line}")
            token = line.split('"')[1]
            if verbose:
                logger.debug(f"token is{token}")
            if token:
                break

    # With the token, get the (3) speed-test-URLS from api.fast.com (which will be in JSON format):
    baseurl = "https://api.fast.com/"
    if force_ipv4:
        # force IPv4 by connecting to an IPv4 address of api.fast.com (over ... HTTP)
        ipv4 = find_ipv4("api.fast.com")
        baseurl = f"http://{ipv4}/"  # HTTPS does not work IPv4 addresses, thus use HTTP
    elif force_ipv6:
        # force IPv6
        ipv6 = find_ipv6("api.fast.com")
        baseurl = f"http://[{ipv6}]/"

    # Not more than 3 possible
    url = f"{baseurl}netflix/speedtest?https=true&token={token}&urlCount=3"
    if verbose:
        logger.debug(f"API url is{url}")
    try:
        url_result = urllib.request.urlopen(url=url, timeout=2)  # 2 second time-out
    except Exception:
        # not good
        if verbose:
            logger.exception(
                "No connection possible"
            )  # probably IPv6, or just no network
        return 0  # no connection, thus no speed

    json_result = url_result.read().decode().strip()
    parsed_json = json.loads(json_result)

    # Prepare for getting those URLs in a threaded way:
    amount = len(parsed_json)
    if verbose:
        logger.debug(f"Number of URLs:{amount}")
    threads = [None] * amount
    results = [0] * amount
    urls = [None] * amount
    for i, json_element in enumerate(parsed_json):
        urls[i] = json_element["url"]  # fill out speed test url from the json format
        if verbose:
            logger.debug(json_element["url"])
    # Let's check whether it's IPv6:
    for url in urls:
        fqdn = url.split("/")[2]
        with contextlib.suppress(Exception):
            socket.getaddrinfo(fqdn, None, socket.AF_INET6)
            if verbose:
                logger.info("IPv6")
    # Now start the threads
    for i, t in enumerate(threads):
        # print("Thread: i is", i)
        t = Thread(target=get_html_result, args=(urls[i], results, i))
        t.daemon = True
        t.start()

    # Monitor the amount of bytes (and speed) of the threads
    time.sleep(1)
    sleep_secs = 3  # 3 seconds sleep
    last_total = 0
    highest_speed_kbps = 0
    nr_loops = int(max_time / sleep_secs)
    for loop in range(nr_loops):
        total = sum(results[i] for i in range(len(threads)))
        delta = total - last_total
        speed_kbps = (delta / sleep_secs) / 1024
        if verbose:
            logger.info(
                f"Loop {loop} Total {total / (1024 * 1024)} MB, "
                f"Delta {delta / (1024 * 1024)} MB, "
                f"Speed {speed_kbps} KB/s "
                f"aka {application_bytes_to_networkbits(speed_kbps) / 1024:.1f} Mbps "
            )
        last_total = total
        if speed_kbps > highest_speed_kbps:
            highest_speed_kbps = speed_kbps
        time.sleep(sleep_secs)

    mbps = round(application_bytes_to_networkbits(highest_speed_kbps) / 1024, 1)
    if verbose:
        logger.info(
            f"Highest Speed (kB/s):{str(highest_speed_kbps)}aka Mbps {str(mbps)}"
        )

    return mbps


if __name__ == "__main__":
    # 	print("let's speed test:")
    # 	print("\nSpeed test, without logging:")
    # 	print(fast_com())
    # 	print("\nSpeed test, with logging:")
    print(fast_com(verbose=True))
    # 	print("\nSpeed test, IPv4, with verbose logging:")
    # 	print(fast_com(verbose=True, max_time=18, force_ipv4=True))
    # 	print("\nSpeed test, IPv6:")
    # 	print(fast_com(max_time=12, force_ipv6=True))
    # 	fast_com(verbose=True, max_time=25)

    # 	print("\ndone")



import base64


def fill_b64(data: str) -> str:
    if len(data) % 4:
        data += "=" * (4 - (len(data) % 4))
    return data


def _url_safe_decode(s: str) -> bytes:
    s = fill_b64(s)
    s = s.replace("-", "+").replace("_", "/")
    return base64.b64decode(s, validate=True)


def encode(s: str) -> bytes:
    return base64.urlsafe_b64encode(s.encode("utf-8"))


def decode(s: str) -> bytes:
    return _url_safe_decode(s)



import json
import signal
import subprocess
from typing import Any, Dict, List, Optional

import aiofiles
from loguru import logger

from ssrspeed.util import PLATFORM


class BaseClient:
    _platform: Optional[str] = PLATFORM

    def __init__(self, clients_dir: str, clients: Dict[str, str], file: str):
        self._clients_dir: str = clients_dir
        self._clients: Dict[str, str] = clients
        self._config_file: str = file
        self._config_list: List[Dict[str, Any]] = []
        self._config_str: str = ""
        self._process: Optional[subprocess.Popen[bytes]] = None
        self._cmd: Dict[str, List[str]] = {}

    async def start_client(self, config: Dict[str, Any], debug: bool = False):
        self._config_str = json.dumps(config)
        async with aiofiles.open(self._config_file, "w+", encoding="utf-8") as f:
            await f.write(self._config_str)

        if self._process is None:
            if BaseClient._platform == "Windows":
                self._process = (
                    subprocess.Popen(self._cmd["win_debug"])
                    if debug
                    else subprocess.Popen(
                        self._cmd["win"],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
                )

                logger.info(
                    f'Starting {self._clients["win"]} with server {config["server"]}:{config["server_port"]}'
                )

            else:
                self._process = (
                    subprocess.Popen(self._cmd["unix"])
                    if debug
                    else subprocess.Popen(
                        self._cmd["unix_debug"],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
                )

                logger.info(
                    f'Starting {self._clients["unix"]} with server {config["server"]}:{config["server_port"]}'
                )

    def check_alive(self) -> bool:
        return self._process.poll() is None  # type: ignore[union-attr]

    # fmt: off
    def stop_client(self):
        if self._process is not None:
            if BaseClient._platform == "Windows":
                self._process.terminate()
            else:
                self._process.send_signal(signal.SIGINT)
        # 	print(self._process.returncode)
            self._process = None
            logger.info("Client terminated.")
    # fmt: on



from . import (
    fast,
    speedtestnet,
    st_asyncio,
    st_netflix as stNF,
    st_socket as stSocket,
    st_stream,
    st_ytb as stYtb,
    webpage_simulation,
)
from .ping import google_ping, tcp_ping

__all__ = [
    "fast",
    "speedtestnet",
    "st_asyncio",
    "st_stream",
    "stNF",
    "stSocket",
    "stYtb",
    "webpage_simulation",
    "google_ping",
    "tcp_ping",
]



import json
import os
import time

import requests
from loguru import logger
from PIL import Image, ImageDraw, ImageFont
from pilmoji import Pilmoji

from ssrspeed.config import ssrconfig
from ssrspeed.result.pusher import push2server
from ssrspeed.result.render import ExporterWps
from ssrspeed.result.sorter import Sorter

TMP_DIR = ssrconfig["path"]["tmp"]
if not os.path.exists(TMP_DIR):
    os.makedirs(TMP_DIR)
TEST_TXT = f"{TMP_DIR}test.txt"
LOGOS_DIR = ssrconfig["path"]["logos"]
RESULTS_DIR = ssrconfig["path"]["results"]
TEMPLATES_DIR = ssrconfig["path"]["templates"]

"""
resultJson
    {
        "group":"GroupName",
        "remarks":"Remarks",
        "loss":0,  # Data loss (0-1)
        "ping":0.014,
        "gping":0.011,
        "dspeed":12435646  # Bytes
        "maxDSpeed":12435646  # Bytes
    }
"""


class ExportResult:
    def __init__(self):
        self.__config: dict = ssrconfig["exportResult"]
        self.__addition: str = ssrconfig["exportResult"]["addition"]
        self.__hide_max_speed: bool = ssrconfig["exportResult"]["hide_max_speed"]
        self.__hide_ntt: bool = not ssrconfig["ntt"]["enabled"]
        self.__hide_geoip: bool = not ssrconfig["geoip"]
        self.__hide_ping: bool = not ssrconfig["ping"]
        self.__hide_gping: bool = not ssrconfig["gping"]
        self.__hide_stream: bool = not ssrconfig["stream"]
        self.__hide_netflix: bool = not ssrconfig["netflix"]
        self.__hide_bilibili = not ssrconfig["bilibili"]
        self.__hide_speed: bool = not ssrconfig["speed"]
        self.__test_method: bool = not ssrconfig["method"]
        self.__hide_stspeed: bool = not ssrconfig["StSpeed"]
        self.__hide_port: bool = not ssrconfig["port"]
        self.__hide_multiplex: bool = not ssrconfig["multiplex"]
        self.__colors: dict = {}
        self.__color_speed_list: list = []
        self.__font: ImageFont.FreeTypeFont = self.set_font(self.__config["font"])
        self.__time_used: str = "N/A"
        self.__upload_config: dict = ssrconfig["uploadResult"]

    # 	self.set_colors()

    def set_hide(self, **kwargs: bool):
        self.__hide_ntt = kwargs.get("ntt", True)
        self.__hide_geoip = kwargs.get("geoip", True)
        self.__hide_ping = kwargs.get("ping", True)
        self.__hide_gping = kwargs.get("gping", self.__hide_gping)
        self.__hide_stream = kwargs.get("stream", True)
        self.__hide_netflix = kwargs.get("netflix", self.__hide_netflix)
        self.__hide_bilibili = kwargs.get("bilibili", self.__hide_bilibili)
        self.__hide_speed = kwargs.get("speed", True)
        self.__hide_stspeed = kwargs.get("StSpeed", self.__hide_stspeed)
        self.__hide_port = kwargs.get("port", True)
        self.__hide_multiplex = kwargs.get("multiplex", True)

    def set_colors(self, name: str = "origin"):
        for color in self.__config["colors"]:
            if color["name"] == name:
                logger.info(f"Set colors as {name}.")
                self.__colors = color["colors"]
                self.__color_speed_list.append(0)
                for speed in self.__colors.keys():
                    try:
                        self.__color_speed_list.append(float(speed))
                    except Exception:
                        continue
                self.__color_speed_list.sort()
                return
        logger.warning(f"Color {name} not found in config.")

    @staticmethod
    def set_font(
        name: str = "SourceHanSansCN-Medium.otf", size: int = 18
    ) -> ImageFont.FreeTypeFont:
        font = ssrconfig["path"]["fonts"] + name
        custom_font = ssrconfig["path"]["custom"] + name
        if os.path.isfile(font):
            return ImageFont.truetype(font, size)
        if os.path.isfile(custom_font):
            return ImageFont.truetype(custom_font, size)
        raise FileNotFoundError(f"Font {name} not found.")

    def set_time_used(self, time_used: float):
        self.__time_used = time.strftime("%H:%M:%S", time.gmtime(time_used))
        logger.info(f"Time Used : {self.__time_used}")

    def export(self, result: list, export_type: int = 0, sort_method: str = ""):
        if not export_type:
            self.__export_as_json(result)
        sorter = Sorter()
        result = sorter.sort_result(result, sort_method)
        self.__export_as_png(result)

    def export_wps_result(self, result: list, export_type: int = 0):
        if not export_type:
            result = self.__export_as_json(result)
        epwps = ExporterWps(result, RESULTS_DIR, TEMPLATES_DIR)
        epwps.export()

    def __get_max_width(self, result: list) -> tuple:
        font = self.__font
        pilmoji = Pilmoji(Image.new("RGB", (1, 1), (255, 255, 255)))
        max_group_width = 0
        max_remark_width = 0
        len_in = 0
        len_out = 0
        for item in result:
            group = item["group"]
            remark = item["remarks"]
            inres = item["InRes"]
            outres = item["OutRes"]
            max_group_width = max(max_group_width, pilmoji.getsize(group, font=font)[0])
            max_remark_width = max(
                max_remark_width, pilmoji.getsize(remark, font=font)[0]
            )
            len_in = max(len_in, pilmoji.getsize(inres, font=font)[0])
            len_out = max(len_out, pilmoji.getsize(outres, font=font)[0])
        return max_group_width + 10, max_remark_width + 10, len_in + 20, len_out + 20

    def __get_base_pos(self, width: float, text: str) -> float:
        font = self.__font
        draw = ImageDraw.Draw(Image.new("RGB", (1, 1), (255, 255, 255)))
        text_size = draw.textsize(text, font=font)[0]
        base_pos = (width - text_size) / 2
        logger.debug(f"Base Position {base_pos}")
        return base_pos

    def __export_as_png(self, result: list):
        if not self.__color_speed_list:
            self.set_colors()
        result_font = self.__font
        generated_time = time.localtime()
        image_height = len(result) * 30 + 30
        weight = self.__get_max_width(result)
        group_width = weight[0]
        remark_width = weight[1]
        in_width = weight[2]
        out_width = weight[3]
        group_width = max(group_width, 60)
        if remark_width < 60:
            remark_width = 90
        in_width = max(in_width, 180)
        out_width = max(out_width, 180)
        other_width = 100

        abema_logo = Image.open(f"{LOGOS_DIR}Abema.png")
        abema_logo.thumbnail((28, 28))
        bahamut_logo = Image.open(f"{LOGOS_DIR}Bahamut.png")
        bahamut_logo.thumbnail((28, 28))
        bilibili_logo = Image.open(f"{LOGOS_DIR}Bilibili.png")
        bilibili_logo.thumbnail((28, 28))
        dazn_logo = Image.open(f"{LOGOS_DIR}Dazn.png")
        dazn_logo.thumbnail((28, 28))
        disney_logo = Image.open(f"{LOGOS_DIR}DisneyPlus.png")
        disney_logo.thumbnail((28, 28))
        hbo_logo = Image.open(f"{LOGOS_DIR}HBO.png")
        hbo_logo.thumbnail((28, 28))
        netflix_logo = Image.open(f"{LOGOS_DIR}Netflix.png")
        netflix_logo.thumbnail((28, 28))
        tvb_logo = Image.open(f"{LOGOS_DIR}TVB.png")
        tvb_logo.thumbnail((28, 28))
        youtube_logo = Image.open(f"{LOGOS_DIR}YouTube.png")
        youtube_logo.thumbnail((28, 28))

        group_right_position = group_width
        remark_right_position = group_right_position + remark_width
        image_right_position = remark_right_position

        if not self.__hide_gping:
            image_right_position = remark_right_position + other_width
        loss_right_position = image_right_position

        if not self.__hide_ping:
            image_right_position = loss_right_position + other_width
        tcp_ping_right_position = image_right_position

        if not self.__hide_gping:
            image_right_position = tcp_ping_right_position + other_width + 25
        google_ping_right_position = image_right_position

        if not self.__hide_port:
            image_right_position = google_ping_right_position + other_width
        port_right_position = image_right_position

        if not self.__hide_speed:
            image_right_position = port_right_position + other_width
        dspeed_right_position = image_right_position

        if not (self.__hide_max_speed or self.__hide_speed):
            image_right_position = dspeed_right_position + other_width
        max_dspeed_right_position = image_right_position

        if not self.__hide_ntt:
            image_right_position = image_right_position + other_width + 80
        ntt_right_position = image_right_position

        if not self.__hide_netflix:
            image_right_position = image_right_position + other_width + 60
        netflix_right_position = image_right_position

        if not self.__hide_bilibili:
            image_right_position = image_right_position + other_width + 60
        bilibili_right_position = image_right_position

        if not self.__hide_stream:
            image_right_position = image_right_position + other_width + 200
        stream_right_position = image_right_position

        if not self.__hide_geoip:
            image_right_position = image_right_position + in_width
        inbound_right_position = image_right_position

        if not self.__hide_geoip:
            image_right_position = image_right_position + out_width
        outbound_right_position = image_right_position

        if not self.__hide_multiplex:
            image_right_position = image_right_position + other_width + 10
        multiplex_right_position = image_right_position

        new_image_height = image_height + 30 * 3
        result_img = Image.new(
            "RGB", (image_right_position, new_image_height), (255, 255, 255)
        )
        draw = ImageDraw.Draw(result_img)
        pilmoji = Pilmoji(result_img, emoji_position_offset=(0, 3))

        # draw.line(
        #     (
        #         0,
        #         new_image_height - 30 - 1,
        #         image_right_position,
        #         new_image_height - 30 - 1,
        #     ),
        #     fill=(127, 127, 127),
        #     width=1,
        # )

        try:
            text = f'\u2708\uFE0F 机场测评图 with SSRSpeed N ( v{ssrconfig["VERSION"]} )'
            pilmoji.text(
                (int(self.__get_base_pos(image_right_position, text)), 4),
                text,
                font=result_font,
                fill=(0, 0, 0),
            )
        except Exception:
            text = f'机场测评图 with SSRSpeed N ( v{ssrconfig["VERSION"]} )'
            draw.text(
                (self.__get_base_pos(image_right_position, text), 4),
                text,
                font=result_font,
                fill=(0, 0, 0),
            )
        draw.line((0, 30, image_right_position - 1, 30), fill=(127, 127, 127), width=1)

        draw.line((1, 0, 1, new_image_height - 1), fill=(127, 127, 127), width=1)
        draw.line(
            (group_right_position, 30, group_right_position, image_height + 30 - 1),
            fill=(127, 127, 127),
            width=1,
        )
        draw.line(
            (remark_right_position, 30, remark_right_position, image_height + 30 - 1),
            fill=(127, 127, 127),
            width=1,
        )

        if not self.__hide_gping:
            draw.line(
                (loss_right_position, 30, loss_right_position, image_height + 30 - 1),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_ping:
            draw.line(
                (
                    tcp_ping_right_position,
                    30,
                    tcp_ping_right_position,
                    image_height + 30 - 1,
                ),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_gping:
            draw.line(
                (
                    google_ping_right_position,
                    30,
                    google_ping_right_position,
                    image_height + 30 - 1,
                ),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_port:
            draw.line(
                (port_right_position, 30, port_right_position, image_height + 30 - 1),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_speed:
            draw.line(
                (
                    dspeed_right_position,
                    30,
                    dspeed_right_position,
                    image_height + 30 - 1,
                ),
                fill=(127, 127, 127),
                width=1,
            )
            if not self.__hide_max_speed:
                draw.line(
                    (
                        max_dspeed_right_position,
                        30,
                        max_dspeed_right_position,
                        image_height + 30 - 1,
                    ),
                    fill=(127, 127, 127),
                    width=1,
                )

        if not self.__hide_ntt:
            draw.line(
                (ntt_right_position, 30, ntt_right_position, image_height + 30 - 1),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_netflix:
            draw.line(
                (
                    netflix_right_position,
                    30,
                    netflix_right_position,
                    image_height + 30 - 1,
                ),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_bilibili:
            draw.line(
                (
                    bilibili_right_position,
                    30,
                    bilibili_right_position,
                    image_height + 30 - 1,
                ),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_stream:
            draw.line(
                (
                    stream_right_position,
                    30,
                    stream_right_position,
                    image_height + 30 - 1,
                ),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_geoip:
            draw.line(
                (
                    inbound_right_position,
                    30,
                    inbound_right_position,
                    image_height + 30 - 1,
                ),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_geoip:
            draw.line(
                (
                    outbound_right_position,
                    30,
                    outbound_right_position,
                    image_height + 30 - 1,
                ),
                fill=(127, 127, 127),
                width=1,
            )

        if not self.__hide_multiplex:
            draw.line(
                (
                    multiplex_right_position,
                    30,
                    multiplex_right_position,
                    image_height + 30 - 1,
                ),
                fill=(127, 127, 127),
                width=1,
            )

        draw.line(
            (image_right_position, 0, image_right_position, new_image_height - 1),
            fill=(127, 127, 127),
            width=1,
        )

        draw.line((0, 0, image_right_position - 1, 0), fill=(127, 127, 127), width=1)

        draw.text(
            (self.__get_base_pos(group_right_position, "Group"), 30 + 4),
            "Group",
            font=result_font,
            fill=(0, 0, 0),
        )

        draw.text(
            (
                group_right_position
                + self.__get_base_pos(
                    remark_right_position - group_right_position, "Remarks"
                ),
                30 + 4,
            ),
            "Remarks",
            font=result_font,
            fill=(0, 0, 0),
        )

        if not self.__hide_gping:
            draw.text(
                (
                    remark_right_position
                    + self.__get_base_pos(
                        loss_right_position - remark_right_position, "Loss"
                    ),
                    30 + 4,
                ),
                "Loss",
                font=result_font,
                fill=(0, 0, 0),
            )

        if not self.__hide_ping:
            draw.text(
                (
                    loss_right_position
                    + self.__get_base_pos(
                        tcp_ping_right_position - loss_right_position, "Ping"
                    ),
                    30 + 4,
                ),
                "Ping",
                font=result_font,
                fill=(0, 0, 0),
            )

        if not self.__hide_gping:
            draw.text(
                (
                    tcp_ping_right_position
                    + self.__get_base_pos(
                        google_ping_right_position - tcp_ping_right_position,
                        "Google Ping",
                    ),
                    30 + 4,
                ),
                "Google Ping",
                font=result_font,
                fill=(0, 0, 0),
            )

        if not self.__hide_port:
            draw.text(
                (
                    google_ping_right_position
                    + self.__get_base_pos(
                        port_right_position - google_ping_right_position, "Port"
                    ),
                    30 + 4,
                ),
                "Port",
                font=result_font,
                fill=(0, 0, 0),
            )

        if not (self.__hide_stspeed or self.__hide_speed):
            if self.__test_method == "NETFLIX":
                draw.text(
                    (
                        port_right_position
                        + self.__get_base_pos(
                            dspeed_right_position - port_right_position, "EndSpeed"
                        ),
                        30 + 4,
                    ),
                    "EndSpeed",
                    font=result_font,
                    fill=(0, 0, 0),
                )
            elif self.__test_method == "YOUTUBE":
                draw.text(
                    (
                        port_right_position
                        + self.__get_base_pos(
                            dspeed_right_position - port_right_position, "StSpeed"
                        ),
                        30 + 4,
                    ),
                    "StSpeed",
                    font=result_font,
                    fill=(0, 0, 0),
                )
            else:
                draw.text(
                    (
                        port_right_position
                        + self.__get_base_pos(
                            dspeed_right_position - port_right_position, "单线程"
                        ),
                        30 + 4,
                    ),
                    "单线程",
                    font=result_font,
                    fill=(0, 0, 0),
                )
        elif not self.__hide_speed:
            if self.__test_method == "NETFLIX":
                draw.text(
                    (
                        port_right_position
                        + self.__get_base_pos(
                            dspeed_right_position - port_right_position, "EndSpeed"
                        ),
                        30 + 4,
                    ),
                    "EndSpeed",
                    font=result_font,
                    fill=(0, 0, 0),
                )
            elif self.__test_method == "YOUTUBE":
                draw.text(
                    (
                        port_right_position
                        + self.__get_base_pos(
                            dspeed_right_position - port_right_position, "StSpeed"
                        ),
                        30 + 4,
                    ),
                    "StSpeed",
                    font=result_font,
                    fill=(0, 0, 0),
                )
            else:
                draw.text(
                    (
                        port_right_position
                        + self.__get_base_pos(
                            dspeed_right_position - port_right_position, "AvgSpeed"
                        ),
                        30 + 4,
                    ),
                    "AvgSpeed",
                    font=result_font,
                    fill=(0, 0, 0),
                )

        if not (self.__hide_max_speed or self.__hide_speed):
            if not self.__hide_stspeed:
                draw.text(
                    (
                        dspeed_right_position
                        + self.__get_base_pos(
                            max_dspeed_right_position - dspeed_right_position, "多线程"
                        ),
                        30 + 4,
                    ),
                    "多线程",
                    font=result_font,
                    fill=(0, 0, 0),
                )
            else:
                draw.text(
                    (
                        dspeed_right_position
                        + self.__get_base_pos(
                            max_dspeed_right_position - dspeed_right_position,
                            "MaxSpeed",
                        ),
                        30 + 4,
                    ),
                    "MaxSpeed",
                    font=result_font,
                    fill=(0, 0, 0),
                )

        if not self.__hide_ntt:
            draw.text(
                (
                    max_dspeed_right_position
                    + self.__get_base_pos(
                        ntt_right_position - max_dspeed_right_position, "UDP NAT Type"
                    ),
                    30 + 4,
                ),
                "UDP NAT Type",
                font=result_font,
                fill=(0, 0, 0),
            )

        if not self.__hide_netflix:
            draw.text(
                (
                    ntt_right_position
                    + self.__get_base_pos(
                        netflix_right_position - ntt_right_position, "Netfilx 解锁"
                    ),
                    30 + 4,
                ),
                "Netfilx 解锁",
                font=result_font,
                fill=(0, 0, 0),
            )

        if not self.__hide_bilibili:
            draw.text(
                (
                    netflix_right_position
                    + self.__get_base_pos(
                        bilibili_right_position - netflix_right_position, "Netfilx 解锁"
                    ),
                    30 + 4,
                ),
                "Bilibili 解锁",
                font=result_font,
                fill=(0, 0, 0),
            )

        if not self.__hide_stream:
            draw.text(
                (
                    bilibili_right_position
                    + self.__get_base_pos(
                        stream_right_position - bilibili_right_position, "流媒体解锁"
                    ),
                    30 + 4,
                ),
                "流媒体解锁",
                font=result_font,
                fill=(0, 0, 0),
            )
        draw.line((0, 60, image_right_position - 1, 60), fill=(127, 127, 127), width=1)

        if not self.__hide_geoip:
            draw.text(
                (
                    stream_right_position
                    + self.__get_base_pos(
                        inbound_right_position - stream_right_position, "Inbound Geo"
                    ),
                    30 + 4,
                ),
                "Inbound Geo",
                font=result_font,
                fill=(0, 0, 0),
            )

        if not self.__hide_geoip:
            draw.text(
                (
                    inbound_right_position
                    + self.__get_base_pos(
                        outbound_right_position - inbound_right_position, "Outbound Geo"
                    ),
                    30 + 4,
                ),
                "Outbound Geo",
                font=result_font,
                fill=(0, 0, 0),
            )

        if not self.__hide_multiplex:
            draw.text(
                (
                    outbound_right_position
                    + self.__get_base_pos(
                        multiplex_right_position - outbound_right_position, "复用检测"
                    ),
                    30 + 4,
                ),
                "复用检测",
                font=result_font,
                fill=(0, 0, 0),
            )

        total_traffic = 0
        online_node = 0
        for i, r in enumerate(result):
            total_traffic += r["trafficUsed"] if (r["trafficUsed"] > 0) else 0
            if (r["ping"] > 0 and r["gPing"] > 0) or (r["dspeed"] > 0):
                online_node += 1

            j = i + 1
            draw.line(
                (0, 30 * j + 60, image_right_position, 30 * j + 60),
                fill=(127, 127, 127),
                width=1,
            )

            group = r["group"]
            remarks = r["remarks"]
            try:
                pilmoji.text(
                    (5, 30 * j + 30 + 4), group, font=result_font, fill=(0, 0, 0)
                )
                pilmoji.text(
                    (group_right_position + 5, 30 * j + 30 + 4),
                    remarks,
                    font=result_font,
                    fill=(0, 0, 0, 0),
                )
            except Exception:
                draw.text((5, 30 * j + 30 + 4), group, font=result_font, fill=(0, 0, 0))
                draw.text(
                    (group_right_position + 5, 30 * j + 30 + 4),
                    remarks,
                    font=result_font,
                    fill=(0, 0, 0, 0),
                )

            if not self.__hide_gping:
                loss = f'{r["gPingLoss"] * 100:.2f}%'
                pos = remark_right_position + self.__get_base_pos(
                    loss_right_position - remark_right_position, loss
                )
                draw.text(
                    (pos, 30 * j + 30 + 4), loss, font=result_font, fill=(0, 0, 0)
                )

            if not self.__hide_ping:
                ping = f'{r["ping"] * 1000:.2f}'
                pos = loss_right_position + self.__get_base_pos(
                    tcp_ping_right_position - loss_right_position, ping
                )
                draw.text(
                    (pos, 30 * j + 30 + 4), ping, font=result_font, fill=(0, 0, 0)
                )

            if not self.__hide_gping:
                g_ping = f'{r["gPing"] * 1000:.2f}'
                pos = tcp_ping_right_position + self.__get_base_pos(
                    google_ping_right_position - tcp_ping_right_position, g_ping
                )
                draw.text(
                    (pos, 30 * j + 30 + 4), g_ping, font=result_font, fill=(0, 0, 0)
                )

            if not self.__hide_port:
                port = str(r["port"])
                pos = google_ping_right_position + self.__get_base_pos(
                    port_right_position - google_ping_right_position, port
                )
                draw.text(
                    (pos, 30 * j + 30 + 4), port, font=result_font, fill=(0, 0, 0)
                )

            if not self.__hide_speed:
                speed = r["dspeed"]
                if speed == -1:
                    pos = port_right_position + self.__get_base_pos(
                        dspeed_right_position - port_right_position, "N/A"
                    )
                    draw.text(
                        (pos, 30 * j + 30 + 1), "N/A", font=result_font, fill=(0, 0, 0)
                    )
                else:
                    draw.rectangle(
                        (
                            port_right_position + 1,
                            30 * j + 30 + 1,
                            dspeed_right_position - 1,
                            30 * j + 60 - 1,
                        ),
                        self.__get_color(speed),
                    )
                    speed = self.__parse_speed(speed)
                    pos = port_right_position + self.__get_base_pos(
                        dspeed_right_position - port_right_position, speed
                    )
                    draw.text(
                        (pos, 30 * j + 30 + 1), speed, font=result_font, fill=(0, 0, 0)
                    )

            if not (self.__hide_max_speed or self.__hide_speed):
                max_speed = r["maxDSpeed"]
                if max_speed == -1:
                    pos = dspeed_right_position + self.__get_base_pos(
                        max_dspeed_right_position - dspeed_right_position, "N/A"
                    )
                    draw.text(
                        (pos, 30 * j + 30 + 1), "N/A", font=result_font, fill=(0, 0, 0)
                    )
                else:
                    draw.rectangle(
                        (
                            dspeed_right_position + 1,
                            30 * j + 30 + 1,
                            max_dspeed_right_position - 1,
                            30 * j + 60 - 1,
                        ),
                        self.__get_color(max_speed),
                    )
                    max_speed = self.__parse_speed(max_speed)
                    pos = dspeed_right_position + self.__get_base_pos(
                        max_dspeed_right_position - dspeed_right_position, max_speed
                    )
                    draw.text(
                        (pos, 30 * j + 30 + 1),
                        max_speed,
                        font=result_font,
                        fill=(0, 0, 0),
                    )

            if not self.__hide_ntt:
                nat_type = r["ntt"]["type"]
                if not nat_type:
                    pos = max_dspeed_right_position + self.__get_base_pos(
                        ntt_right_position - max_dspeed_right_position, "Unknown"
                    )
                    draw.text(
                        (pos, 30 * j + 30 + 1),
                        "Unknown",
                        font=result_font,
                        fill=(0, 0, 0),
                    )
                else:
                    pos = max_dspeed_right_position + self.__get_base_pos(
                        ntt_right_position - max_dspeed_right_position, nat_type
                    )
                    draw.text(
                        (pos, 30 * j + 30 + 1),
                        nat_type,
                        font=result_font,
                        fill=(0, 0, 0),
                    )

            if not self.__hide_netflix:
                netflix_type = r["Ntype"]
                pos = ntt_right_position + self.__get_base_pos(
                    netflix_right_position - ntt_right_position, netflix_type
                )
                draw.text(
                    (pos, 30 * j + 30 + 1),
                    netflix_type,
                    font=result_font,
                    fill=(0, 0, 0),
                )

            if not self.__hide_bilibili:
                bilibili_type = r["Bltype"]
                pos = netflix_right_position + self.__get_base_pos(
                    bilibili_right_position - netflix_right_position, bilibili_type
                )
                draw.text(
                    (pos, 30 * j + 30 + 1),
                    bilibili_type,
                    font=result_font,
                    fill=(0, 0, 0),
                )

            if not self.__hide_stream:
                netflix_type = r["Ntype"]
                hbo_type = r["Htype"]
                disney_type = r["Dtype"]
                youtube_type = r["Ytype"]
                abema_type = r["Atype"]
                bahamut_type = r["Btype"]
                dazn_type = r["Dztype"]
                tvb_type = r["Ttype"]
                bilibili_type = r["Bltype"]
                n_type = bool(netflix_type[:4] == "Full")
                bl_type = bool(bilibili_type != "N/A")
                sums = (
                    n_type
                    + hbo_type
                    + disney_type
                    + youtube_type
                    + abema_type
                    + bahamut_type
                    + dazn_type
                    + bl_type
                    + tvb_type
                )
                pos = (
                    bilibili_right_position
                    + (stream_right_position - bilibili_right_position - sums * 35) / 2
                )
                if n_type:
                    result_img.paste(netflix_logo, (int(pos), 30 * j + 30 + 1))
                    pos += 35
                if hbo_type:
                    result_img.paste(hbo_logo, (int(pos), 30 * j + 30 + 1))
                    pos += 35
                if disney_type:
                    result_img.paste(disney_logo, (int(pos), 30 * j + 30 + 1))
                    pos += 35
                if youtube_type:
                    result_img.paste(youtube_logo, (int(pos), 30 * j + 30 + 1))
                    pos += 35
                if abema_type:
                    result_img.paste(abema_logo, (int(pos), 30 * j + 30 + 1))
                    pos += 35
                if bahamut_type:
                    result_img.paste(bahamut_logo, (int(pos), 30 * j + 30 + 1))
                    pos += 35
                if dazn_type:
                    result_img.paste(dazn_logo, (int(pos), 30 * j + 30 + 1))
                    pos += 35
                if tvb_type:
                    result_img.paste(tvb_logo, (int(pos), 30 * j + 30 + 1))
                    pos += 35
                if bl_type:
                    result_img.paste(bilibili_logo, (int(pos), 30 * j + 30 + 1))
                    pos += 35

            if not self.__hide_geoip:
                inbound_geo = r["InRes"]
                pos = stream_right_position + self.__get_base_pos(
                    inbound_right_position - stream_right_position, inbound_geo
                )
                draw.text(
                    (pos, 30 * j + 30 + 1),
                    inbound_geo,
                    font=result_font,
                    fill=(0, 0, 0),
                )

            if not self.__hide_geoip:
                outbound_geo = r["OutRes"]
                pos = inbound_right_position + self.__get_base_pos(
                    outbound_right_position - inbound_right_position, outbound_geo
                )
                draw.text(
                    (pos, 30 * j + 30 + 1),
                    outbound_geo,
                    font=result_font,
                    fill=(0, 0, 0),
                )

            if not self.__hide_geoip:
                inbound_ip = r["InIP"]
                outbound_ip = r["OutIP"]
                multiplex_res = ""
                if outbound_ip != "N/A":
                    inbound_mul = -1
                    outbound_mul = -1
                    if inbound_ip == "N/A":
                        inbound_mul = 0
                    for _, ipitem in enumerate(result):
                        if ipitem["InIP"] == inbound_ip and inbound_ip != "N/A":
                            inbound_mul += 1
                        if ipitem["OutIP"] == outbound_ip:
                            outbound_mul += 1
                    if inbound_mul and outbound_mul:
                        multiplex_res = "完全复用"
                    elif (not inbound_mul) and (not outbound_mul):
                        multiplex_res = "无复用"
                    elif inbound_mul:
                        multiplex_res = "中转复用"
                    elif outbound_mul:
                        multiplex_res = "落地复用"
                else:
                    multiplex_res = "未知"

                pos = outbound_right_position + self.__get_base_pos(
                    multiplex_right_position - outbound_right_position, multiplex_res
                )
                draw.text(
                    (pos, 30 * j + 30 + 1),
                    multiplex_res,
                    font=result_font,
                    fill=(0, 0, 0),
                )

        files = []
        if total_traffic < 0:
            traffic_used = "N/A"
        else:
            traffic_used = self.__parse_traffic(total_traffic)

        if not self.__hide_speed:
            t1 = f"Traffic used : {traffic_used}. "
        else:
            t1 = ""

        if not self.__hide_gping:
            t2 = f" Online Node(s) : [{online_node}/{len(result)}]"
        else:
            t2 = ""

        with open(TEST_TXT, "a+", encoding="utf-8") as test:
            test.seek(0)
            url = test.readline()
            try:
                sum0 = int(test.readline())
            except ValueError:
                sum0 = 0
        os.remove(TEST_TXT)

        if not self.__hide_speed:
            clash_ua = {"User-Agent": "Clash"}

            try:
                r = requests.get(url, headers=clash_ua, timeout=15)
                t = r.headers["subscription-userinfo"]
                dl = int(t[t.find("download") + 9 : t.find("total") - 2])
                sum_ = dl
            except Exception:
                sum_ = 0
            avgrate = (sum_ - sum0) / total_traffic if total_traffic else 0
            if (sum_ - sum0) > 0:
                t3 = f".  AvgRate : {avgrate:.2f}"
            else:
                t3 = ""
        else:
            t3 = ""

        draw.text(
            (5, image_height + 30 + 4),
            t1 + f"Time used: {self.__time_used}." + t2 + t3,
            font=result_font,
            fill=(0, 0, 0),
        )

        # draw.line(
        #     (
        #         0,
        #         new_image_height - 30 * 3 - 1,
        #         image_right_position,
        #         new_image_height - 30 * 3 - 1,
        #     ),
        #     fill=(127, 127, 127),
        #     width=1,
        # )

        draw.text(
            (5, image_height + 30 * 2 + 4),
            f'{self.__addition}  Generated at {time.strftime("%Y-%m-%d %H:%M:%S", generated_time)}',
            font=result_font,
            fill=(0, 0, 0),
        )
        draw.line(
            (
                0,
                new_image_height - 30 - 1,
                image_right_position,
                new_image_height - 30 - 1,
            ),
            fill=(127, 127, 127),
            width=1,
        )

        # draw.line(
        #     (
        #         0,
        #         new_image_height - 30 - 1,
        #         image_right_position,
        #         new_image_height - 30 - 1,
        #     ),
        #     fill=(127, 127, 127),
        #     width=1,
        # )
        # draw.text(
        #     (5, image_height + 30 * 2 + 4),
        #     f'By SSRSpeed {ssrconfig["VERSION"]}.',
        #     font=result_font,
        #     fill=(0, 0, 0),
        # )

        draw.line(
            (0, new_image_height - 1, image_right_position, new_image_height - 1),
            fill=(127, 127, 127),
            width=1,
        )
        filename = (
            RESULTS_DIR + time.strftime("%Y-%m-%d-%H-%M-%S", generated_time) + ".png"
        )
        result_img.save(filename)
        files.append(filename)
        logger.info(f"Result image saved as {filename}")

        if not self.__config.get("uploadResult", False):
            self.__upload_result(files)

    def __upload_result(self, files: list):
        sever = self.__upload_config.get("server", "")
        token = self.__upload_config.get("token", "")
        remark = self.__upload_config.get("remark", "")
        for _file in files:
            push2server(_file, sever, token, remark)

    @staticmethod
    def __parse_traffic(traffic: float) -> str:
        traffic = traffic / 1024 / 1024
        if traffic < 1:
            return f"{traffic * 1024:.2f} KB"
        gb_traffic = traffic / 1024
        return f"{traffic:.2f} MB" if gb_traffic < 1 else f"{gb_traffic:.2f} GB"

    @staticmethod
    def __parse_speed(speed: float) -> str:
        speed = speed / 1024 / 1024
        return f"{speed * 1024:.2f} KB" if speed < 1 else f"{speed:.2f} MB"

    @staticmethod
    def __new_mix_color(lc: dict, rc: dict, rt: int) -> tuple:
        #   print("RGB1 : {lc}, RGB2 : {rc}, RT : {rt}")
        return (
            int(lc[0] * (1 - rt) + rc[0] * rt),
            int(lc[1] * (1 - rt) + rc[1] * rt),
            int(lc[2] * (1 - rt) + rc[2] * rt),
        )

    def __get_color(self, data: float) -> tuple:
        if not self.__color_speed_list:
            return 255, 255, 255
        cur_speed = self.__color_speed_list[-1]
        back_speed = 0
        if data >= cur_speed * 1024 * 1024:
            return (
                self.__colors[str(cur_speed)][0],
                self.__colors[str(cur_speed)][1],
                self.__colors[str(cur_speed)][2],
            )
        for i, s in enumerate(self.__color_speed_list):
            cur_speed = s * 1024 * 1024
            if i > 0:
                back_speed = self.__color_speed_list[i - 1]
            back_speed_str = str(back_speed)
            #   print(f"{data / 1024 / 1024} {back_speed}")
            if data < cur_speed:
                rgb1 = (
                    self.__colors[back_speed_str] if back_speed > 0 else (255, 255, 255)
                )
                rgb2 = self.__colors[str(s)]
                rt = (data - back_speed * 1024 * 1024) / (
                    cur_speed - back_speed * 1024 * 1024
                )
                logger.debug(
                    f"Speed : {data / 1024 / 1024}, RGB1 : {rgb1}, RGB2 : {rgb2}, RT : {rt}"
                )
                return self.__new_mix_color(rgb1, rgb2, rt)
        return 255, 255, 255

    @staticmethod
    def __export_as_json(result: list) -> list:
        filename = (
            RESULTS_DIR + time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime()) + ".json"
        )
        with open(filename, "w+", encoding="utf-8") as f:
            f.writelines(
                json.dumps(result, sort_keys=True, indent=4, separators=(",", ":"))
            )
        logger.info(f"Result exported as {filename}")
        return result



from typing import Optional

from loguru import logger

from ssrspeed.util import b64plus


class ParserV2RayQuantumult:
    @staticmethod
    def parse_subs_config(raw_link: str) -> Optional[dict]:
        link = raw_link[8:]
        link_decoded = b64plus.decode(link).decode("utf-8")
        try:
            link_split = link_decoded.split(",")
            new_list = []
            while True:
                try:
                    item = link_split.pop(0)
                except IndexError:
                    break
                text = item
                while text.count('"') % 2 != 0:
                    try:
                        text += f", {link_split.pop(0)}"
                    except IndexError as error:
                        raise ValueError("Invalid Quantumult URL.") from error
                new_list.append(text)
            link_split = new_list

            server = link_split[1]
            remarks = link_split[0].split(" = ")[0] or server
            port = int(link_split[2])
            security = link_split[3]
            uuid = link_split[4].replace('"', "")
            group = link_split[5].split("=")[1]

            tls = ""
            tls_host = ""
            host = ""  # http host, web socket host, h2 host, quic encrypt method
            net = "tcp"
            path = ""  # Websocket path, http path, quic encrypt key
            headers: list = []
            allow_insecure = True
            i = 7
            if link_split[6].split("=")[1] == "true":
                tls = "tls"
                tls_host = link_split[7].split("=")[1]
                allow_insecure = link_split[8].split("=")[1] != "1"
                i = 8
            if len(link_split) in {11, 12}:
                net = link_split[i + 1].split("=")[1]
                path = link_split[i + 2].split("=")[1].replace('"', "")
                if header := (
                    link_split[i + 3].split("=")[1].replace('"', "").split("[Rr][Nn]")
                ):
                    host = header[0].split(":")[1].strip()
                    headers.extend(
                        {
                            "header": header[h].split(":")[0].strip(),
                            "value": header[h].split(":")[1].strip(),
                        }
                        for h in range(1, len(header))
                    )

            _type = "none"  # Obfs type under tcp mode
            aid = 0
            logger.debug(
                f"Server : {server}, Port : {port}, tls-host : {tls_host}, Path : {path}, "
                f"Type : {_type}, UUID : {uuid}, AlterId : {aid}, Network : {net}, "
                f"Host : {host}, Headers : {headers}, TLS : {tls}, Remarks : {remarks}, "
                f"group={group}"
            )
            return {
                "remarks": remarks,
                "group": group,
                "server": server,
                "server_port": port,
                "id": uuid,
                "alterId": aid,
                "security": security,
                "allowInsecure": allow_insecure,
                "type": _type,
                "path": path,
                "network": net,
                "host": host,
                "headers": headers,
                "tls": tls,
                "tls-host": tls_host,
            }

        except Exception:
            logger.exception(f"Parse {raw_link} failed. (Quantumult Method)")
            return None


if __name__ == "__main__":
    LINK = "vmess://T3JhY2xlID0gdm1lc3MsIHYycmF5LnVydXRvcmE5Ni50aywgNDQzLCBjaGFjaGEyMC1pZXRmLXBvbHkxMzA1LCAiNjZiZTk4YTctNTUzNy00ZjM5LWIwYjgtY2FhNGU3ZGI0ZmQwIiwgZ3JvdXA9VjJSYXlQcm92aWRlciwgb3Zlci10bHM9dHJ1ZSwgdGxzLWhvc3Q9djJyYXkudXJ1dG9yYTk2LnRrLCBjZXJ0aWZpY2F0ZT0xLCBvYmZzPXdzLCBvYmZzLXBhdGg9Ii91cnV0b3JhOTYiLCBvYmZzLWhlYWRlcj0iSG9zdDogdjJyYXkudXJ1dG9yYTk2LnRrIg"
    print(ParserV2RayQuantumult.parse_subs_config(LINK))



import asyncio
import re

import aiohttp
from aiohttp_socks import ProxyConnector
from loguru import logger

# Netflix requestIpAddress regex compile
nf_ip_re = re.compile(r'"requestIpAddress":"(.*)"')


class StreamTest:
    @classmethod
    async def netflix(cls, host, headers, inner_dict, port, outbound_ip):
        logger.info(f"Performing netflix test LOCAL_PORT: {port}.")
        try:
            sum_ = 0
            async with aiohttp.ClientSession(
                headers=headers,
                connector=ProxyConnector(host=host, port=port),
                timeout=aiohttp.ClientTimeout(connect=10),
            ) as session, session.get(
                url="https://www.netflix.com/title/70242311"
            ) as response1:
                netflix_ip = "N/A"
                if response1.status == 200:
                    sum_ += 1
                    netflix_ip = nf_ip_re.findall(str(await response1.read()))[0].split(
                        ","
                    )[0]
                    logger.info(f"Netflix IP : {netflix_ip}")
                async with session.get(
                    url="https://www.netflix.com/title/70143836"
                ) as response2:
                    rg = ""
                    if response2.status == 200:
                        sum_ += 1
                        rg = f"({str(response2.url).split('com/')[1].split('/')[0]})"
                    if rg == "(title)":
                        rg = "(us)"
                    # 测试连接状态
                    if sum_ == 0:
                        logger.info("Netflix test result: None.")
                        inner_dict["Ntype"] = "None"
                    elif sum_ == 1:
                        logger.info("Netflix test result: Only Original.")
                        inner_dict["Ntype"] = "Only Original"
                    elif outbound_ip == netflix_ip:
                        logger.info("Netflix test result: Full Native.")
                        inner_dict["Ntype"] = f"Full Native{rg}"
                    else:
                        logger.info("Netflix test result: Full DNS.")
                        inner_dict["Ntype"] = f"Full DNS{rg}"
        except Exception as e:
            logger.error(f"Netflix error: {str(e)}")
            return {}

    @classmethod
    async def hbomax(cls, host, headers, inner_dict, port):
        logger.info(f"Performing HBO max test LOCAL_PORT: {port}.")
        try:
            async with aiohttp.ClientSession(
                headers=headers,
                connector=ProxyConnector(host=host, port=port),
                timeout=aiohttp.ClientTimeout(connect=10),
            ) as session, session.get(
                url="https://www.hbomax.com/", allow_redirects=False
            ) as response:
                inner_dict["Htype"] = response.status == 200
        except Exception as e:
            logger.error(f"HBO max error: {str(e)}")

    @classmethod
    async def disneyplus(cls, host, headers, inner_dict, port):
        logger.info(f"Performing Disney plus test LOCAL_PORT: {port}.")
        try:
            async with aiohttp.ClientSession(
                headers=headers,
                connector=ProxyConnector(host=host, port=port),
                timeout=aiohttp.ClientTimeout(connect=5),
            ) as session, session.get(
                url="https://www.disneyplus.com/"
            ) as response1, session.get(
                url="https://global.edge.bamgrid.com/token"
            ) as response2:
                if response1.status == 200 and response2.status != 403:
                    text = await response1.text()
                    if text.find("Region", 0, 400) == -1:
                        inner_dict["Dtype"] = False
                    elif response1.history:
                        if 300 <= response1.history[0].status <= 399:
                            inner_dict["Dtype"] = False
                    else:
                        inner_dict["Dtype"] = True
                else:
                    inner_dict["Dtype"] = False
        except Exception as e:
            logger.error(f"Disney plus error: {str(e)}")

    @classmethod
    async def youtube(cls, host, headers, inner_dict, port):
        logger.info(f"Performing Youtube Premium test LOCAL_PORT: {port}.")
        try:
            async with aiohttp.ClientSession(
                headers=headers,
                connector=ProxyConnector(host=host, port=port),
                timeout=aiohttp.ClientTimeout(connect=10),
            ) as session, session.get(
                url="https://music.youtube.com/", allow_redirects=False
            ) as response:
                if "is not available" in await response.text():
                    inner_dict["Ytype"] = False
                elif response.status == 200:
                    inner_dict["Ytype"] = True
                else:
                    inner_dict["Ytype"] = False
        except Exception as e:
            logger.error(f"Youtube Premium error: {str(e)}")

    @classmethod
    async def abema(cls, host, headers, inner_dict, port):
        logger.info(f"Performing Abema test LOCAL_PORT: {port}.")
        try:
            async with aiohttp.ClientSession(
                headers=headers,
                connector=ProxyConnector(host=host, port=port),
                timeout=aiohttp.ClientTimeout(connect=10),
            ) as session, session.get(
                url="https://api.abema.io/v1/ip/check?device=android",
                allow_redirects=False,
            ) as response:
                text = await response.text()
                inner_dict["Atype"] = text.count("Country") > 0
        except Exception as e:
            logger.error(f"Abema error: {str(e)}")

    @classmethod
    async def bahamut(cls, host, headers, inner_dict, port):
        logger.info(f"Performing Bahamut test LOCAL_PORT: {port}.")
        try:
            async with aiohttp.ClientSession(
                headers=headers,
                connector=ProxyConnector(host=host, port=port),
                timeout=aiohttp.ClientTimeout(connect=10),
            ) as session, session.get(
                url="https://ani.gamer.com.tw/ajax/token.php?adID=89422&sn=14667",
                allow_redirects=False,
            ) as response:
                text = await response.text()
                inner_dict["Btype"] = text.count("animeSn") > 0
        except Exception as e:
            logger.error(f"Bahamut error: {str(e)}")

    @classmethod
    async def indazn(cls, host, headers, inner_dict, port):
        logger.info(f"Performing Dazn test LOCAL_PORT: {port}.")
        try:
            async with aiohttp.ClientSession(
                headers=headers,
                connector=ProxyConnector(host=host, port=port, verify_ssl=False),
                timeout=aiohttp.ClientTimeout(connect=10),
            ) as session:
                payload = {
                    "LandingPageKey": "generic",
                    "Languages": "zh-CN,zh,en",
                    "Platform": "web",
                    "PlatformAttributes": {},
                    "Manufacturer": "",
                    "PromoCode": "",
                    "Version": "2",
                }
                async with session.post(
                    url="https://startup.core.indazn.com/misl/v5/Startup",
                    json=payload,
                    allow_redirects=False,
                ) as response:
                    inner_dict["Dztype"] = response.status == 200
        except Exception as e:
            logger.error(f"Dazn error: {str(e)}")

    @classmethod
    async def mytvsuper(cls, host, headers, inner_dict, port):
        logger.info(f"Performing TVB test LOCAL_PORT: {port}.")
        try:
            async with aiohttp.ClientSession(
                headers=headers,
                connector=ProxyConnector(host=host, port=port),
                timeout=aiohttp.ClientTimeout(connect=10),
            ) as session, session.get(
                url="https://www.mytvsuper.com/iptest.php", allow_redirects=False
            ) as response:
                text = await response.text()
                inner_dict["Ttype"] = text.count("HK") > 0
        except Exception as e:
            logger.error(f"TVB error: {str(e)}")

    @classmethod
    async def bilibili(cls, host, headers, inner_dict, port):
        logger.info(f"Performing Bilibili test LOCAL_PORT: {port}.")
        try:
            async with aiohttp.ClientSession(
                headers=headers,
                connector=ProxyConnector(host=host, port=port),
                timeout=aiohttp.ClientTimeout(connect=10),
            ) as session:
                params = {
                    "avid": 50762638,
                    "cid": 100279344,
                    "qn": 0,
                    "type": "",
                    "otype": "json",
                    "ep_id": 268176,
                    "fourk": 1,
                    "fnver": 0,
                    "fnval": 16,
                    "module": "bangumi",
                }
                async with session.get(
                    url="https://api.bilibili.com/pgc/player/web/playurl",
                    params=params,
                    allow_redirects=False,
                ) as response:
                    if response.status == 200:
                        json_data = await response.json()
                        if json_data["code"] == 0:
                            inner_dict["Bltype"] = "台湾"
                        else:
                            params = {
                                "avid": 18281381,
                                "cid": 29892777,
                                "qn": 0,
                                "type": "",
                                "otype": "json",
                                "ep_id": 183799,
                                "fourk": 1,
                                "fnver": 0,
                                "fnval": 16,
                                "module": "bangumi",
                            }
                            session.cookie_jar.clear()
                            async with session.get(
                                url="https://api.bilibili.com/pgc/player/web/playurl",
                                params=params,
                                allow_redirects=False,
                            ) as response2:
                                if response2.status == 200:
                                    json_data2 = await response2.json()
                                    if json_data2["code"] == 0:
                                        inner_dict["Bltype"] = "港澳台"
                                else:
                                    inner_dict["Bltype"] = "N/A"
        except Exception as e:
            logger.error(f"Bilibili error: {str(e)}")


async def start_stream_test(port, stream_cfg, outbound_ip):
    host = "127.0.0.1"
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36"
    }
    test_list = []
    inner_dict = {
        "Ntype": "None",
        "Htype": False,
        "Dtype": False,
        "Ytype": False,
        "Atype": False,
        "Btype": False,
        "Dztype": False,
        "Ttype": False,
        "Bltype": "N/A",
    }
    if stream_cfg["HBO_TEST"]:
        test_list.append(
            asyncio.create_task(StreamTest.hbomax(host, headers, inner_dict, port))
        )
    if stream_cfg["DISNEY_TEST"]:
        test_list.append(
            asyncio.create_task(StreamTest.disneyplus(host, headers, inner_dict, port))
        )
    if stream_cfg["YOUTUBE_TEST"]:
        test_list.append(
            asyncio.create_task(StreamTest.youtube(host, headers, inner_dict, port))
        )
    if stream_cfg["ABEMA_TEST"]:
        test_list.append(
            asyncio.create_task(StreamTest.abema(host, headers, inner_dict, port))
        )
    if stream_cfg["BAHAMUT_TEST"]:
        test_list.append(
            asyncio.create_task(StreamTest.bahamut(host, headers, inner_dict, port))
        )
    if stream_cfg["DAZN_TEST"]:
        test_list.append(
            asyncio.create_task(StreamTest.indazn(host, headers, inner_dict, port))
        )
    if stream_cfg["TVB_TEST"]:
        test_list.append(
            asyncio.create_task(StreamTest.mytvsuper(host, headers, inner_dict, port))
        )
    if stream_cfg["BILIBILI_TEST"]:
        test_list.append(
            asyncio.create_task(StreamTest.bilibili(host, headers, inner_dict, port))
        )
    if stream_cfg["NETFLIX_TEST"]:
        test_list.append(
            asyncio.create_task(
                StreamTest.netflix(host, headers, inner_dict, port, outbound_ip)
            )
        )
    await asyncio.wait(test_list)
    return inner_dict


if __name__ == "__main__":
    STREAM_CFG = {
        "NETFLIX_TEST": True,
        "HBO_TEST": True,
        "DISNEY_TEST": True,
        "YOUTUBE_TEST": True,
        "ABEMA_TEST": True,
        "BAHAMUT_TEST": True,
        "DAZN_TEST": True,
        "TVB_TEST": True,
        "BILIBILI_TEST": True,
    }
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    logger.info(asyncio.run(start_stream_test(7890, STREAM_CFG, "N/A")))



import asyncio
import copy
import socket
import threading
import time

import socks
from loguru import logger

from ssrspeed.util import ip_loc
from ssrspeed.util.rule import DownloadRuleMatch

DEFAULT_SOCKET = socket.socket
MAX_TIME = 0
TOTAL_RECEIVED = 0
EXIT_FLAG = False
LOCK = threading.Lock()
MAX_FILE_SIZE = 100 * 1024 * 1024


def restore_socket():
    socket.socket = DEFAULT_SOCKET


def speed_test_thread(link, buffer):
    global TOTAL_RECEIVED, MAX_TIME
    logger.debug(f"Thread {threading.current_thread().ident} started.")
    link = link.replace("https://", "").replace("http://", "")
    host = link[: link.find("/")]
    request_uri = link[link.find("/") :]
    logger.debug(f"\nLink: {link}\nHost: {host}\nRequestUri: {request_uri}")

    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(12)
        try:
            s.connect((host, 80))
            logger.debug(f"Connected to {host}")
        except Exception:
            logger.error(f"Connect to {host} error.")
            with LOCK:
                TOTAL_RECEIVED += 0
            return None
        s.send(
            b"GET %b HTTP/1.1\r\nHost: %b\r\nUser-Agent: Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
            b"AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.121 Safari/537.36\r\n\r\n "
            % (request_uri.encode("utf-8"), host.encode("utf-8"))
        )
        logger.debug("Request sent.")
        start_time = time.time()
        received = 0
        while True:
            try:
                xx = s.recv(buffer)
            except socket.timeout:
                logger.error("Receive data timeout.")
                break
            except ConnectionResetError:
                logger.warning("Remote host closed connection.")
                break
            lxx = len(xx)
            received += lxx
            with LOCK:
                TOTAL_RECEIVED += lxx
            if received >= MAX_FILE_SIZE or EXIT_FLAG:
                break
        end_time = time.time()
        delta_time = end_time - start_time
        if delta_time >= 12:
            delta_time = 11
        s.close()
        logger.debug(
            f"Thread {threading.current_thread().ident} done,time : {delta_time}"
        )
        with LOCK:
            MAX_TIME = max(MAX_TIME, delta_time)
    except Exception:
        logger.exception("")
        return 0


async def speed_test_socket(
    file_download: dict, address: str, port: int, speed_test: bool, st_speed_test: bool
) -> tuple:
    if not speed_test:
        return 0, 0, [], 0
    workers = file_download.get("workers", 4)
    buffer_ = file_download.get("buffer", 4096)
    avg_st_speed: float = 0
    global EXIT_FLAG, MAX_TIME, TOTAL_RECEIVED, MAX_FILE_SIZE

    dlrm = DownloadRuleMatch(file_download)
    res = dlrm.get_url(await ip_loc(port))

    MAX_FILE_SIZE = res[1] * 1024 * 1024
    MAX_TIME = 0
    TOTAL_RECEIVED = 0
    EXIT_FLAG = False
    socks.set_default_proxy(socks.SOCKS5, address, port)
    socket.socket = socks.socksocket  # type: ignore

    if st_speed_test:
        for _ in range(1):
            nmsl = threading.Thread(target=speed_test_thread, args=(res[0], buffer_))
            nmsl.start()

        max_speed_list = []
        old_received = 0
        for i in range(1, 11):
            await asyncio.sleep(0.5)
            with LOCK:
                delta_received = TOTAL_RECEIVED - old_received
                old_received = TOTAL_RECEIVED
            current_speed = delta_received / 0.5
            max_speed_list.append(current_speed)
            print(
                "\r["
                + "=" * i
                + f"> [{i * 10}%/100%] [{current_speed / 1024 / 1024:.2f} MB/s]",
                end="",
            )
            if EXIT_FLAG:
                print(
                    "\r["
                    + "=" * i
                    + f"] [100%/100%] [{current_speed / 1024 / 1024:.2f} MB/s]"
                )
                break
        EXIT_FLAG = True
        for _ in range(10):
            time.sleep(0.1)
            if MAX_TIME != 0:
                break
        if MAX_TIME == 0:
            logger.error("Socket Test Error !")
            return 0, 0, [], 0

        logger.info(
            f"SingleThread: Fetched {TOTAL_RECEIVED / 1024:.2f} KB in {MAX_TIME:.2f} s."
        )

        avg_st_speed = TOTAL_RECEIVED / MAX_TIME

        MAX_TIME = 0
        TOTAL_RECEIVED = 0
        EXIT_FLAG = False

    for _ in range(workers):
        nmsl = threading.Thread(target=speed_test_thread, args=(res[0], buffer_))
        nmsl.start()

    max_speed_list = []
    current_speed = 0
    old_received = 0
    for i in range(1, 11):
        time.sleep(0.5)
        with LOCK:
            delta_received = TOTAL_RECEIVED - old_received
            old_received = TOTAL_RECEIVED
        current_speed = delta_received / 0.5
        max_speed_list.append(current_speed)
        print(
            "\r["
            + "=" * i
            + f"> [{i * 10}%/100%] [{current_speed / 1024 / 1024:.2f} MB/s]",
            end="",
        )
        if EXIT_FLAG:
            print(
                "\r["
                + "=" * i
                + f"] [100%/100%] [{current_speed / 1024 / 1024:.2f} MB/s]"
            )
            break
    EXIT_FLAG = True
    for _ in range(10):
        time.sleep(0.1)
        if MAX_TIME != 0:
            break
    if MAX_TIME == 0:
        logger.error("Socket Test Error !")
        return 0, 0, [], 0

    restore_socket()
    raw_speed_list = copy.deepcopy(max_speed_list)
    max_speed_list.sort()
    if len(max_speed_list) > 7:
        msum = sum(max_speed_list[i] for i in range(7, len(max_speed_list) - 2))
        max_speed = msum / (len(max_speed_list) - 2 - 7)
    else:
        max_speed = current_speed
    logger.info(
        f"MultiThread: Fetched {TOTAL_RECEIVED / 1024:.2f} KB in {MAX_TIME:.2f} s."
    )
    avg_speed = TOTAL_RECEIVED / MAX_TIME

    if not st_speed_test:
        avg_st_speed = avg_speed
        avg_speed = max_speed

    return avg_st_speed, avg_speed, raw_speed_list, TOTAL_RECEIVED


if __name__ == "__main__":
    from ssrspeed.path import ROOT_PATH, get_path_json

    key_path = get_path_json(ROOT_PATH)
    from ssrspeed.config import generate_config_file, load_path_config, ssrconfig

    load_path_config({"path": key_path})
    generate_config_file()
    logger.info(
        asyncio.run(
            speed_test_socket(ssrconfig["fileDownload"], "127.0.0.1", 7890, True, True)
        )
    )
